#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import logging

class BaseNotification(object):

    _config = {}

    _EVENTS = None

    @classmethod
    def update_config(cls, new_config):
        cls._config.update(new_config)

    @classmethod
    def register_eventlog_events(cls, events):
        cls._EVENTS = events

    @classmethod
    def set_redis_con(cls, r):
        cls.__redis_conn = r

    @classmethod
    def _get_subject(cls, alert, custom_message=None):
        """
        >>> BaseNotification._get_subject({'is_alert': True, 'changed': True, 'alert_def':{'name': 'Test'}, 'entity':{'id':'hostxy'}, 'captures': {}})
        'NEW ALERT: Test on hostxy'

        >>> BaseNotification._get_subject({'is_alert': True, 'changed': False, 'alert_def':{'name': 'Test'}, 'entity':{'id':'hostxy'}, 'captures': {}, 'duration': datetime.timedelta(seconds=30)})
        'ALERT ONGOING: Test on hostxy for 0:00:30'
        """

        if alert['changed']:
            event = ('NEW ALERT' if alert and alert.get('is_alert') else 'ALERT ENDED')
        else:
            event = 'ALERT ONGOING'

        name = cls._get_expanded_alert_name(alert, custom_message)

        if not custom_message:
            return ('{}: {} on {} for {}'.format(event, name, alert['entity']['id'], str(alert['duration'
                    ])[:7]) if alert.get('duration') else '{}: {} on {}'.format(event, name, alert['entity']['id']))
        else:
            return '{}: {}'.format(event, name)

    @classmethod
    def _get_expanded_alert_name(cls, alert, custom_message=None):
        name = (alert['alert_def']['name'] if not custom_message else custom_message)
        try:
            replacements = {'entities': alert['entity']['id']}
            replacements.update(alert['captures'])
            return name.format(**replacements)
        except KeyError, e:
            return name  # This is fairly normal. Just use the unformatted name.
        except Exception, e:
            return "<<< Unformattable name '{name}': {message} >>>".format(name=name, message=e)

    @classmethod
    def send(cls, alert, *args, **kwargs):
        raise NotImplementedError('Method meant to be overriden by subclass')

    @classmethod
    def resolve_group(cls, targets, phone=False):
        new_targets = []
        for target in targets:
            prefix = target[0:target.find(':')+1]
            if not prefix in ['group:', 'active:']:
                new_targets.append(target)
                continue

            group_id = target[target.find(':')+1:]

            key = 'zmon:group:'+group_id + (':members' if prefix == 'group:' else ':active')
            team = cls.__redis_conn.smembers(key)

            if not team:
                logging.warn("no members found for group: %s", target)
                continue

            if not phone:
                new_targets.extend(team)
            else:
                for m in team:
                    phone_numbers = cls.__redis_conn.smembers('zmon:member:'+m+':phone')
                    new_targets.extend(phone_numbers)

        logging.info("Redirect notifications: from %s to %s", targets, new_targets)
        return new_targets




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Zalando-specific function to query EventLog
"""

from zmon_worker_monitor.zmon_worker.errors import CheckError
#from http import HttpWrapper  # FIXME: watch out for this!!!

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial
from zmon_worker_monitor import plugin_manager


class EventlogFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(EventlogFactory, self).__init__()
        # fields from configuration
        self.eventlog_url = None
        # fields from dependencies: plugin depends 1 other plugin
        self.http_factory = None

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        self.eventlog_url = conf['eventlog_url']

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """

        # load plugins dependencies and store them locally for efficiency
        if not self.http_factory:
            self.http_factory = plugin_manager.get_plugin_obj_by_name('http', 'Function')

        return propartial(EventLogWrapper,
                          http_wrapper=self.http_factory.create(factory_ctx),
                          url=self.eventlog_url)


class EventLogWrapper(object):

    '''Convenience wrapper to access EventLog counts'''

    def __init__(self, http_wrapper, url):
        self.__http = http_wrapper
        self.url = url.rstrip('/') + '/'

    def __request(self, path, params):
        return self.__http(self.url + path, params=params).json()

    def count(self, event_type_ids, time_from, time_to=None, group_by=None, **kwargs):
        '''Return number of events for given type IDs in given time frame

        returns a single number (integer) if only one type ID is given
        returns a dict (typeId as hex=>count) if more than one type ID is given
        returns a dict (fieldValue => count) if one type ID is given and a field name with "group_by"

        >>> EventLogWrapper(object, 'https://eventlog.example.com/').count('a', time_from='-1h')
        Traceback (most recent call last):
            ...
        CheckError: EventLog type ID must be a integer

        >>> EventLogWrapper(object, 'https://eventlog.example.com/').count(123, time_from='-1h')
        Traceback (most recent call last):
            ...
        CheckError: EventLog type ID is out of range
        '''

        if isinstance(event_type_ids, (int, long)):
            event_type_ids = [event_type_ids]
        for type_id in event_type_ids:
            if not isinstance(type_id, (int, long)):
                raise CheckError('EventLog type ID must be a integer')
            if type_id < 0x1001 or type_id > 0xfffff:
                raise CheckError('EventLog type ID is out of range')
        params = kwargs
        params['event_type_ids'] = ','.join(['{:x}'.format(v) for v in event_type_ids])
        params['time_from'] = time_from
        params['time_to'] = time_to
        params['group_by'] = group_by
        result = self.__request('count', params)
        if len(event_type_ids) == 1 and not group_by:
            return result.get(params['event_type_ids'], 0)
        else:
            return result


if __name__ == '__main__':

    import sys
    import logging
    logging.basicConfig(level=logging.DEBUG)

    # init plugin manager and collect plugins, as done by Zmon when worker is starting
    plugin_manager.init_plugin_manager()
    plugin_manager.collect_plugins(load_builtins=True, load_env=True)

    eventlog_url = sys.argv[1]
    factory_ctx = {}

    http = plugin_manager.get_plugin_obj_by_name('http', 'Function').create(factory_ctx)

    #eventlog = EventLogWrapper()
    eventlog = EventLogWrapper(http_wrapper=http, url=eventlog_url)

    print eventlog.count(0x96001, time_from='-1m')
    print eventlog.count([0x96001, 0x63005], time_from='-1m')
    print eventlog.count(0x96001, time_from='-1m', group_by='appDomainId')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import requests
import logging

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial

logger = logging.getLogger('zmon-worker.scalyr-function')

class ScalyrWrapperFactory(IFunctionFactoryPlugin):
    def __init__(self):
        super(ScalyrWrapperFactory, self).__init__()

    def configure(self, conf):
        self.read_key = conf.get('read.key', '')
        return

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(ScalyrWrapper, read_key=self.read_key)


class ScalyrWrapper(object):

    def __init__(self, read_key):
        self.numeric_url = 'https://www.scalyr.com/api/numericQuery'
        self.timeseries_url = 'https://www.scalyr.com/api/timeseriesQuery'
        self.facet_url = 'https://www.scalyr.com/api/facetQuery'
        self.read_key = read_key

    def count(self, query, minutes=5):

        val = {
            'token': self.read_key,
            'queryType': 'numeric',
            'filter': query,
            'function': 'count',
            'startTime': str(minutes)+'m',
            'priority': 'low',
            'buckets': 1
        }

        r = requests.post(self.numeric_url, data=json.dumps(val), headers={"Content-Type": "application/json"})
        j = r.json()
        if 'values' in j:
            return j['values'][0]
        else:
            return j

    def function(self, function, query, minutes=5):

        val = {
            'token': self.read_key,
            'queryType': 'numeric',
            'filter': query,
            'function': function,
            'startTime': str(minutes)+'m',
            'priority': 'low',
            'buckets': 1
        }

        r = requests.post(self.numeric_url, data=json.dumps(val), headers={"Content-Type": "application/json"})
        j = r.json()
        if 'values' in j:
            return j['values'][0]
        else:
            return j

    def facets(self, filter, field, max_count=5, minutes=30, prio="low"):

        val = {
            'token': self.read_key,
            'queryType': 'facet',
            'filter': filter,
            'field': field,
            'maxCount': max_count,
            "startTime": str(minutes)+"m",
            "priority": prio
        }

        r = requests.post(self.facet_url, data=json.dumps(val), headers={"Content-Type": "application/json"})
        j = r.json()
        return j


    def timeseries(self, filter, function="count", minutes=30, buckets=1, prio="low"):

        val = {
            'token': self.read_key,
            'queries': [
                {
                    "filter": filter,
                    "function": function,
                    "startTime": str(minutes)+"m",
                    "buckets": buckets,
                    "priority": prio
                }
            ]
        }

        r = requests.post(self.timeseries_url, data=json.dumps(val), headers={"Content-Type": "application/json"})
        j = r.json()
        if j['status'] == 'success':
            if len(j['results'][0]['values'])==1:
                return j['results'][0]['values'][0]
            return map(lambda x: x * minutes / buckets, j['results'][0]['values'])
        return j


if __name__ == '__main__':
    import os
    s = ScalyrWrapper(read_key=os.getenv('SCALYR_READ_KEY'))
    print s.count(query="$application_id='zmon-scheduler'")




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
from zmon_worker_monitor.zmon_worker.errors import DbError

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial


# default port Oracle Net Listener port
DEFAULT_PORT = 1521
MAX_RESULTS = 100


class SqlOracleFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(SqlOracleFactory, self).__init__()
        # fields from config
        self._user = None
        self._pass = None

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        self._user = conf['user']
        self._pass = conf['pass']

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(SqlOracleWrapper, factory_ctx['host'], factory_ctx['port'], factory_ctx['entity'].get('sid'),
                          user=self._user, password=self._pass)


def _check_oracle_env():
    if 'ORACLE_HOME' not in os.environ or 'LD_LIBRARY_PATH' not in os.environ or os.environ['ORACLE_HOME'] \
        not in os.environ['LD_LIBRARY_PATH'] or not os.path.isdir(os.environ['ORACLE_HOME']):
        raise Exception('Environmet variables ORACLE_HOME and LD_LIBRARY_PATH are not correctly set')


def _import_db_driver():
    try:
        _check_oracle_env()
        _cx_Oracle = __import__('cx_Oracle', globals(), locals(), [], -1)
    except Exception:
        # do we want to take some action here?
        raise
    return _cx_Oracle


class SqlOracleWrapper(object):

    '''Oracle SQL adapter
    sql().execute('SELECT 1').result()
    '''

    def __init__(self, host, port, sid, user='nagios', password='', enabled=True):
        self._stmt = None
        self._dsn_tns = None
        self._enabled = enabled
        self.__cx_Oracle = None
        self.__conn = None
        self.__cursor = None

        if self._enabled:
            self.__cx_Oracle = _import_db_driver()
            port = (int(port) if port and str(port).isdigit() else DEFAULT_PORT)
            try:
                self._dsn_tns = self.__cx_Oracle.makedsn(host, port, sid)
                self.__conn = self.__cx_Oracle.connect(user, password, self._dsn_tns)
                self.__cursor = self.__conn.cursor()
            except Exception, e:
                raise DbError(str(e), operation='Connect to dsn={}'.format(self._dsn_tns)), None, sys.exc_info()[2]

    def execute(self, stmt):
        self._stmt = stmt
        return self

    def result(self, agg=sum):
        '''return single row result, will result primitive value if only one column is selected'''

        result = {}
        try:
            if self._enabled and self.__cx_Oracle:
                cur = self.__cursor
                try:
                    cur.execute(self._stmt)
                    desc = [d[0] for d in cur.description]  # Careful: col names come out all uppercase
                    row = cur.fetchone()
                    if row:
                        result = dict(zip(desc, row))
                finally:
                    cur.close()
        except Exception, e:
            raise DbError(str(e), operation=self._stmt), None, sys.exc_info()[2]

        if len(result) == 1:
            return result.values()[0]
        else:
            return result

    def results(self):
        '''return many rows'''

        results = []
        try:
            if self._enabled and self.__cx_Oracle:
                cur = self.__cursor
                try:
                    cur.execute(self._stmt)
                    desc = [d[0] for d in cur.description]  # Careful: col names come out all uppercase
                    rows = cur.fetchmany(MAX_RESULTS)
                    for row in rows:
                        row = dict(zip(desc, row))
                        results.append(row)
                finally:
                    cur.close()
        except Exception, e:
            raise DbError(str(e), operation=self._stmt), None, sys.exc_info()[2]
        return results


if __name__ == '__main__':

    default_sql_stmt = "SELECT 'OK' from dual"

    if len(sys.argv) in (6, 7):
        check = SqlOracleWrapper(sys.argv[1], sys.argv[2], sys.argv[3], user=sys.argv[4], password=sys.argv[5])
        if len(sys.argv) == 7:
            sql_stmt = sys.argv[6]
        else:
            print 'executing default statement:', default_sql_stmt
            sql_stmt = default_sql_stmt

        # print '>>> one result:\n', check.execute(sql_stmt).result()
        print '>>> many results:\n', check.execute(sql_stmt).results()
    else:
        print '{} <host> <port> <sid> [sql_stmt]'.format(sys.argv[0])




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import logging
from zmon_worker_monitor.zmon_worker.errors import DbError

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial


logger = logging.getLogger(__name__)

# requires:
# sudo apt-get install freetds-dev
# sudo pip install pymssql

# default port
DEFAULT_PORT = 1433
MAX_RESULTS = 100


class MsSqlFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(MsSqlFactory, self).__init__()
        # fields from config
        self._user = None
        self._pass = None

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        self._user = conf['user']
        self._pass = conf['pass']

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(MsSqlWrapper, factory_ctx['host'], factory_ctx['port'], factory_ctx['database'],
                          user=self._user, password=self._pass, timeout=factory_ctx['soft_time_limit'])


def _import_db_driver():
    try:
        _cx_MsSql = __import__('pymssql', globals(), locals(), [], -1)
    except Exception, e:
        logger.exception('Import of module pymssql failed')
        raise
    return _cx_MsSql


class MsSqlWrapper(object):

    # Note: Timeout must be in seconds
    def __init__(self, host, port, database, user='robot_zmon', password='', enabled=True, timeout=60):
        self.__stmt = None
        self.__enabled = enabled

        if self.__enabled:
            self.__cx_MsSql = _import_db_driver()
            port = (int(port) if port and str(port).isdigit() else DEFAULT_PORT)
            try:
                self.__conn = self.__cx_MsSql.connect('{}:{}'.format(host, port), user, password, database, timeout,
                                                      as_dict=True)
                self.__cursor = self.__conn.cursor()
            except Exception, e:
                raise DbError(str(e), operation='Connect to {}:{}'.format(host, port)), None, sys.exc_info()[2]

    def execute(self, stmt):
        self.__stmt = stmt
        return self

    def result(self):
        # return single row result, will result primitive value if only one column is selected

        result = {}
        try:
            if self.__enabled and self.__cx_MsSql:
                cur = self.__cursor
                try:
                    cur.execute(self.__stmt)
                    row = cur.fetchone()
                    if row:
                        result = row
                finally:
                    cur.close()
        except Exception, e:
            raise DbError(str(e), operation=self.__stmt), None, sys.exc_info()[2]

        if len(result) == 1:
            return result.values()[0]
        else:
            return result

    def results(self):
        # return many rows

        results = []
        try:
            if self.__enabled and self.__cx_MsSql:
                cur = self.__cursor
                try:
                    cur.execute(self.__stmt)
                    rows = cur.fetchmany(MAX_RESULTS)
                    for row in rows:
                        results.append(row)
                finally:
                    cur.close()
        except Exception, e:
            raise DbError(str(e), operation=self.__stmt), None, sys.exc_info()[2]
        return results


if __name__ == '__main__':

    logging.basicConfig(level=logging.DEBUG)

    default_sql_stmt = "SELECT 'ONE ROW' AS COLUMN_NAME"

    if len(sys.argv) in (6, 7):
        check = MsSqlWrapper(sys.argv[1], sys.argv[2], sys.argv[3], user=sys.argv[4], password=sys.argv[5])
        if len(sys.argv) == 7:
            sql_stmt = sys.argv[6]
        else:
            print 'executing default statement:', default_sql_stmt
            sql_stmt = default_sql_stmt

        print '>>> one result:\n', check.execute(sql_stmt).result()
        # print '>>> many results:\n', check.execute(sql_stmt).results()
    else:
        print '{} <host> <port> <database> [sql_stmt]'.format(sys.argv[0])


from notification import BaseNotification

import requests
import logging
logger = logging.getLogger(__name__)

class NotifySlack(BaseNotification):

    @classmethod
    def send(cls, alert, *args, **kwargs):
        url = "https://slack.com/api/chat.postMessage"
        token = kwargs.get('token', cls._config.get('notifications.slack.token'))
        repeat = kwargs.get('repeat', 0)

        message = {"as_user":"true", "token": token, "channel": kwargs.get('channel', '#general'), "text": kwargs.get("message", cls._get_subject(alert))}

        try:
            logger.info("Sending to %s %s", url, message)
            r = requests.post(url, params=message, verify=False)
            r.raise_for_status()
        except Exception as ex:
            logger.exception("Slack write failed %s", ex)

        return repeat



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Server module for exposing an rpc interface for clients to remotely control a local ProcessManager
"""

import os
import sys
import signal
import settings
import logging

if __name__ == '__main__':
    import logging.config
    logging.config.dictConfig(settings.RPC_SERVER_CONF['LOGGING'])

logger = logging.getLogger(__name__)

from pprint import pformat

from process_controller import ProcessController
import worker
import rpc_utils


def save_pid(abort_pidfile=False):
    pid = os.getpid()
    pid_file = os.path.join(settings.data_dir, 'rpc_server.pid')
    if abort_pidfile and os.path.isfile(pid_file):
        print >>sys.stderr, 'pid file {} already exists. Is another process running? Aborting!'.format(pid_file)
        sys.exit(1)
    with open(pid_file, 'w') as f:
        f.write(str(pid))


def sigterm_handler(signum, frame):
    # this will propagate the SystemExit exception all around, so we can quit listening loops, cleanup and exit
    sys.exit(0)


class ProcessControllerProxy(rpc_utils.RpcProxy):
    """
    A proxy class to expose some methods of multiprocess manager server to listen to remote requests,
    possible request are:
     1. start N more child processes
     2. terminate processes with pid1, pid2, ...
     3. report running statistics
     4. report status of process with pid1
     5. terminate all child processes
     6. terminate yourself
    """

    exposed_obj_class = ProcessController

    valid_methods = ['spawn_many', 'list_running', 'list_stats', 'start_action_loop', 'stop_action_loop',
                     'is_action_loop_running', 'get_dynamic_num_processes', 'set_dynamic_num_processes',
                     'get_action_policy', 'set_action_policy', 'available_action_policies', 'terminate_all_processes',
                     'terminate_process', 'mark_for_termination']

    def list_running(self):
        return pformat(self.get_exposed_obj().list_running())

    def list_stats(self):
        return pformat(self.get_exposed_obj().list_stats())

    def on_exit(self):
        self.get_exposed_obj().terminate_all_processes()  # TODO: Think why exit codes are sometimes -15 and others 0


class MainProcess(object):

    def __init__(self):
        save_pid()
        signal.signal(signal.SIGTERM, sigterm_handler)

    def start_proc_control(self):
        self.proc_control = ProcessController(default_target=worker.start_worker, action_policy='report')

    def start_rpc_server(self):

        rpc_proxy = ProcessControllerProxy(self.proc_control)

        rpc_utils.start_RPC_server(settings.RPC_SERVER_CONF['HOST'],
                                   settings.RPC_SERVER_CONF['PORT'],
                                   settings.RPC_SERVER_CONF['RPC_PATH'],
                                   rpc_proxy)


def main(config=None):

    save_pid(abort_pidfile=True)

    signal.signal(signal.SIGTERM, sigterm_handler)

    proc_control = ProcessController(default_target=worker.start_worker,
                                     action_policy='report', always_add_kwargs={'external_config': config})

    rpc_proxy = ProcessControllerProxy(proc_control)

    rpc_utils.start_RPC_server(settings.RPC_SERVER_CONF['HOST'],
                               settings.RPC_SERVER_CONF['PORT'],
                               settings.RPC_SERVER_CONF['RPC_PATH'],
                               rpc_proxy)


if __name__ == '__main__':

    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pythonwhois

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial


class WhoisFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(WhoisFactory, self).__init__()

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        return

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(WhoisWrapper, host=factory_ctx['host'])


class WhoisWrapper(object):

    def __init__(self, host, timeout=10):
        self.host = host
        self.timeout = timeout

    def check(self):
        parsed = {}
        data, server_list = pythonwhois.net.get_whois_raw(self.host, with_server_list=True)
        if len(server_list) > 0:
            parsed = pythonwhois.parse.parse_raw_whois(data, normalized=True, never_query_handles=False,
                                                       handle_server=server_list[-1])
        else:
            parsed = pythonwhois.parse.parse_raw_whois(data, normalized=True)

        return parsed


if __name__ == '__main__':
    import json
    import sys
    import datetime


    def json_fallback(obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        else:
            return obj


    data = WhoisWrapper(sys.argv[1]).check()
    print json.dumps(data, default=json_fallback, indent=4)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import redis
import time

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial


# round to microseconds
ROUND_SECONDS_DIGITS = 6


class CounterFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(CounterFactory, self).__init__()

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        return

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(CounterWrapper, key_prefix='{}:{}:'.format(factory_ctx['check_id'], factory_ctx['entity_id']),
                          redis_host=factory_ctx['redis_host'], redis_port=factory_ctx['redis_port'])


class CounterWrapper(object):

    '''Measure increasing counts (per second) by saving the last value in Redis'''

    def __init__(self, key, redis_host, redis_port=6379, key_prefix=''):
        self.__con = redis.StrictRedis(redis_host, redis_port)
        self.key_prefix = key_prefix
        self.key(key)

    def key(self, key):
        '''expose key setter to allow reusing redis connection (CounterWrapper instance)'''

        self.__key = 'zmon:counters:{}{}'.format(self.key_prefix, key)
        # return self to allow method chaining
        return self

    def per_second(self, value):
        '''return increment rate of counter value (per second)'''

        olddata = self.__con.get(self.__key)
        result = 0
        now = time.time()
        if olddata:
            olddata = json.loads(olddata)
            time_diff = now - olddata['ts']
            value_diff = value - olddata['value']
            # do not allow negative values (important for JMX counters which will reset after restart/deploy)
            result = max(value_diff / time_diff, 0)
        newdata = {'value': value, 'ts': round(now, ROUND_SECONDS_DIGITS)}
        self.__con.set(self.__key, json.dumps(newdata))
        return result

    def per_minute(self, value):
        '''convenience method: returns per_second(..) / 60'''

        return self.per_second(value) / 60.0


if __name__ == '__main__':
    counter = CounterWrapper('test', 'localhost', 6379)
    print counter.per_second(1)
    time.sleep(2)
    print counter.per_second(101)
    time.sleep(1)
    print counter.per_second(111)
    time.sleep(1)
    print counter.per_minute(211)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Zalando-specific function to query DeployCtl job information
"""

from itertools import groupby
from operator import itemgetter

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial
from zmon_worker_monitor import plugin_manager


class JobsFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(JobsFactory, self).__init__()
        # fields from dependencies: plugin depends 1 other plugin
        self.http_factory = None

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        return

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """

        # load plugins dependencies and store them locally for efficiency
        if not self.http_factory:
            self.http_factory = plugin_manager.get_plugin_obj_by_name('http', 'Function')

        return propartial(JobsWrapper,
                          http_wrapper=self.http_factory.create(factory_ctx),
                          project=factory_ctx['entity'].get('name'))


class JobsWrapper(object):

    def __init__(self, http_wrapper, environment, project, **kwargs):
        self.url = 'https://deployctl.example.com/jobs/history.json/{}/{}'.format(environment, project)
        self.__http = http_wrapper
        self.http_wrapper_params = kwargs
        self.name = itemgetter('name')

    def __request(self):
        return self.__http(self.url, **self.http_wrapper_params).json()

    def lastruns(self):
        start_time = itemgetter('start_seconds_ago')

        return dict((job, min(runs, key=start_time)) for (job, runs) in groupby(sorted(self.__request(),
                    key=self.name), key=self.name))

    def history(self):
        return dict((job, list(runs)) for (job, runs) in groupby(sorted(self.__request(), key=self.name),
                    key=self.name))





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import requests
import json
import sys

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial

logger = logging.getLogger('zmon-worker.kairosdb-function')

class KairosdbFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(KairosdbFactory, self).__init__()

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        return

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(KairosdbWrapper, url=factory_ctx.get('entity_url'))

class KairosdbWrapper(object):

    def __init__(self, url):
        self.url = url

    def query(self, name, group_by = [], tags = None, start = -5, end = 0, time_unit='seconds', aggregators = None):
        url = self.url + '/api/v1/datapoints/query'
        q = {
            "start_relative": {
                "value": start,
                "unit": time_unit
            },
            "metrics": [{
                "name": name,
            }]
        }

        if aggregators is not None:
            q["metrics"][0]["aggregators"] = aggregators

        if tags is not None:
            q["metrics"][0]["tags"] = tags

        try:
            response = requests.post(url, json=q)
            if response.status_code == requests.codes.ok:
                return response.json()["queries"][0]
            else:
                raise Exception("KairosDB Query failed: " + json.dumps(q))
        except requests.Timeout, e:
            raise HttpError('timeout', self.url), None, sys.exc_info()[2]
        except requests.ConnectionError, e:
            raise HttpError('connection failed', self.url), None, sys.exc_info()[2]

    def tagnames(self):
        return []

    def tagnames(self):
        return []

    def metric_tags(self):
        return {}


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import redis

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial
from zmon_worker_monitor import plugin_manager


STATISTIC_GAUGE_KEYS = frozenset([
    'blocked_clients',
    'connected_clients',
    'connected_slaves',
    'instantaneous_ops_per_sec',
    'used_memory',
    'used_memory_rss',
])
STATISTIC_COUNTER_KEYS = frozenset([
    'evicted_keys',
    'expired_keys',
    'keyspace_hits',
    'keyspace_misses',
    'total_commands_processed',
    'total_connections_received',
])


class RedisFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(RedisFactory, self).__init__()
        # fields to store dependencies: plugin depends on 1 other plugin
        self.counter_factory = None

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        return

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        # load plugins dependencies and store them locally for efficiency
        if not self.counter_factory:
            self.counter_factory = plugin_manager.get_plugin_obj_by_name('counter', 'Function')

        return propartial(RedisWrapper, counter=self.counter_factory.create(factory_ctx), host=factory_ctx['host'])


class RedisWrapper(object):

    '''Class to allow only readonly access to underlying redis connection'''

    def __init__(self, counter, host, port=6379, db=0, socket_connect_timeout=1):
        self._counter = counter('')
        self.__con = redis.StrictRedis(host, port, db, socket_connect_timeout=socket_connect_timeout)

    def llen(self, key):
        return self.__con.llen(key)

    def lrange(self, key, start, stop):
        return self.__con.lrange(key, start, stop)

    def get(self, key):
        return self.__con.get(key)

    def hget(self, key, field):
        return self.__con.hget(key, field)

    def hgetall(self, key):
        return self.__con.hgetall(key)

    def statistics(self):
        '''
        Return general Redis statistics such as operations/s

        Example result::

            {
                "blocked_clients": 2,
                "commands_processed_per_sec": 15946.48,
                "connected_clients": 162,
                "connected_slaves": 0,
                "connections_received_per_sec": 0.5,
                "dbsize": 27351,
                "evicted_keys_per_sec": 0.0,
                "expired_keys_per_sec": 0.0,
                "instantaneous_ops_per_sec": 29626,
                "keyspace_hits_per_sec": 1195.43,
                "keyspace_misses_per_sec": 1237.99,
                "used_memory": 50781216,
                "used_memory_rss": 63475712
            }
        '''

        data = self.__con.info()
        stats = {}
        for key in STATISTIC_GAUGE_KEYS:
            stats[key] = data.get(key)
        for key in STATISTIC_COUNTER_KEYS:
            stats['{}_per_sec'.format(key).replace('total_', '')] = \
                round(self._counter.key(key).per_second(data.get(key, 0)), 2)
        stats['dbsize'] = self.__con.dbsize()
        return stats


if __name__ == '__main__':
    import sys
    import json

    # init plugin manager and collect plugins, as done by Zmon when worker is starting
    plugin_manager.init_plugin_manager()
    plugin_manager.collect_plugins(load_builtins=True, load_env=True)

    factory_ctx = {
        'redis_host': 'localhost',
    }
    counter = plugin_manager.get_plugin_obj_by_name('counter', 'Function').create(factory_ctx)
    wrapper = RedisWrapper(counter, sys.argv[1])
    print json.dumps(wrapper.statistics(), indent=4, sort_keys=True)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from zmon_worker_monitor.zmon_worker.errors import SnmpError
from pysnmp.entity.rfc3413.oneliner import cmdgen
from pysnmp.proto.rfc1902 import Integer, OctetString, Counter32, Counter64
import re
import subprocess32

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial


class SnmpFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(SnmpFactory, self).__init__()

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        return

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(SnmpWrapper, host=factory_ctx['host'])


class SnmpWrapper(object):

    def __init__(self, host, community='public', version='v2c', timeout=5):
        if re.match(r'^[0-9]+$', str(timeout)):
            self.timeout = timeout
        if re.match(r'^[\w\-.]+$', host):
            self.host = host
        self.generator = cmdgen.CommandGenerator()
        self.transport = cmdgen.UdpTransportTarget((host, 161), timeout=timeout)
        if version == 'v2c':
            # self.comm_data = cmdgen.CommunityData('server', community, 1)  # 1 means version SNMP v2c
            self.comm_data = cmdgen.CommunityData(community)

    def memory(self):
        '''return {'free': 3, ..}
        # 3 UCD-SNMP-MIB::memTotalSwap.0 = INTEGER: 4194300 kB
        # 4 UCD-SNMP-MIB::memAvailSwap.0 = INTEGER: 4194300 kB
        # 5 UCD-SNMP-MIB::memTotalReal.0 = INTEGER: 32816032 kB
        # 6 UCD-SNMP-MIB::memAvailReal.0 = INTEGER: 9392600 kB
        # 11 UCD-SNMP-MIB::memTotalFree.0 = INTEGER: 13586900 kB
        # 12 UCD-SNMP-MIB::memMinimumSwap.0 = INTEGER: 16000 kB
        # 13 UCD-SNMP-MIB::memShared.0 = INTEGER: 0 kB
        # 14 UCD-SNMP-MIB::memBuffer.0 = INTEGER: 299992 kB
        # 15 UCD-SNMP-MIB::memCached.0 = INTEGER: 6115656 kB
        '''

        result = {}
        oids = {
            'swap_total': '1.3.6.1.4.1.2021.4.3.0',
            'swap_free': '1.3.6.1.4.1.2021.4.4.0',
            'ram_total': '1.3.6.1.4.1.2021.4.5.0',
            'ram_free': '1.3.6.1.4.1.2021.4.6.0',
            'ram_total_free': '1.3.6.1.4.1.2021.4.11.0',
            'swap_min': '1.3.6.1.4.1.2021.4.12.0',
            'ram_shared': '1.3.6.1.4.1.2021.4.13.0',
            'ram_buffer': '1.3.6.1.4.1.2021.4.14.0',
            'ram_cache': '1.3.6.1.4.1.2021.4.15.0',
        }
        for k, oid in oids.items():
            val = self._get_cmd(oid)
            result[k] = self.parse(Integer, int, val)
        return result

    def load(self):
        '''
        Return CPU Load Averages
        Example result: {"load1":0.95,"load5":0.69,"load15":0.72}
        '''

        result = {}
        oids = {'load1': '1.3.6.1.4.1.2021.10.1.3.1', 'load5': '1.3.6.1.4.1.2021.10.1.3.2',
                'load15': '1.3.6.1.4.1.2021.10.1.3.3'}
        for k, oid in oids.items():
            val = self._get_cmd(oid)
            result[k] = self.parse(OctetString, lambda x: float(str(x)), val)
        return result

    def cpu(self):
        '''
        Returns CPU Percentage

        Example result: {"cpu_system":0,"cpu_user":19,"cpu_idle":79}
        '''

        result = {}
        oids = {'cpu_idle': '1.3.6.1.4.1.2021.11.11.0', 'cpu_user': '1.3.6.1.4.1.2021.11.9.0',
                'cpu_system': '1.3.6.1.4.1.2021.11.10.0'}
        for k, oid in oids.items():
            val = self._get_cmd(oid)
            result[k] = self.parse(Integer, int, val)
        return result

    def cpu_raw(self):
        result = {}
        oids = {
            'raw_cpu_user': '1.3.6.1.4.1.2021.11.50.0',
            'raw_cpu_system': '.1.3.6.1.4.1.2021.11.52.0',
            'raw_cpu_nice': '.1.3.6.1.4.1.2021.11.51.0',
            'raw_cpu_idle': '.1.3.6.1.4.1.2021.11.53.0',
        }
        for k, oid in oids.items():
            val = self._get_cmd(oid)
            result[k] = self.parse(Counter32, int, val)
        return result

    def df(self):
        '''
        Return disk usage information

        Example result: {"/data/postgres-wal-nfs-itr":{"percentage_space_used":0,"used_space":160,"total_size":524288000,"device":"zala0-1-stp-02:/vol/zal_pgwal","available_space":524287840,"percentage_inodes_used":0}}
        '''

        # disk_table = '.1.3.6.1.4.1.2021.9'
#        base = '1.3.6.1.4.1.2021.13.15.1.1'
        base = '1.3.6.1.4.1.2021.9.1'
        base_idx = base + '.1.'
        base_name = base + '.2.'
        results = {}
        idx2name = {}
        result_all = self._get_walk(base)
        for oid in sorted(result_all.keys()):
            if base_idx in oid:
                val = str(result_all[oid])
                idx2name[val] = None
            elif base_name in oid:
                idx = oid.split('.')[-1]
                val = result_all[oid]
                idx2name[idx] = val
        for oid in sorted(result_all.keys()):
            if oid in base_idx or oid in base_name:
                continue
            idx = oid.split('.')[-1]
            name = str(idx2name[idx])
            if 'loop' in name or 'ram' in name:
                continue
            results[name] = results.get(name, {})
            tname = oid
            kind = oid.split('.')[-2]
            if kind == '3':
                tname = 'device'
            elif kind == '6':
                tname = 'total_size'  # kBytes
            elif kind == '7':
                tname = 'available_space'  # kBytes
            elif kind == '8':
                tname = 'used_space'  # kBytes
            elif kind == '9':
                tname = 'percentage_space_used'
            elif kind == '10':
                tname = 'percentage_inodes_used'
            elif kind == '11':
                tname = 'total_size'  # Gauge32
            elif kind == '13':
                tname = 'available_space'  # Gauge32
            elif kind == '15':
                tname = 'used_space'  # Gauge32
            else:
                continue
            results[name][tname] = result_all[oid]
        return results

    def logmatch(self):
        # logmatch_table = '.1.3.6.1.4.1.2021.16'
        base = '1.3.6.1.4.1.2021.16.2.1'
        base_idx = base + '.1.'
        base_name = base + '.2.'
        results = {}
        idx2name = {}
        result_all = self._get_walk(base)
        for oid in sorted(result_all.keys()):
            if base_idx in oid:
                val = str(result_all[oid])
                idx2name[val] = None
            elif base_name in oid:
                idx = oid.split('.')[-1]
                val = result_all[oid]
                idx2name[idx] = val
        for oid in sorted(result_all.keys()):
            if oid in base_idx or oid in base_name:
                continue
            idx = oid.split('.')[-1]
            name = str(idx2name[idx])
            results[name] = results.get(name, {})
            tname = oid
            kind = oid.split('.')[-2]
            if kind == '3':
                tname = 'file'  # file name
            elif kind == '4':
                tname = 'regex'  # regex string
            elif kind == '5':
                tname = 'global_count'  # counter32
            elif kind == '7':
                tname = 'current_count'  # counter32 since last logrotation
            elif kind == '9':
                tname = 'last_count'  # counter32 since last read
            else:
                continue
            results[name][tname] = result_all[oid]
        return results

    def interfaces(self):
        # IF-MIB::interfaces_table = '1.3.6.1.2.1.2.'
        # IF-MIB::ifMIB_table = '1.3.6.1.2.1.31'
        base = '1.3.6.1.2.1.2.2.1'
        base_idx = base + '.1.'
        base_name = base + '.2.'
        results = {}
        idx2name = {}
        result_all = self._get_walk(base)
        for oid in sorted(result_all.keys()):
            if base_idx in oid:
                val = str(result_all[oid])
                idx2name[val] = None
            elif base_name in oid:
                idx = oid.split('.')[-1]
                val = result_all[oid]
                idx2name[idx] = val
        for oid in sorted(result_all.keys()):
            if oid in base_idx or oid in base_name:
                continue
            idx = oid.split('.')[-1]
            name = str(idx2name[idx])
            results[name] = results.get(name, {})
            tname = oid
            kind = oid.split('.')[-2]
            if kind == '7':
                tname = 'opStatus'  # operational status
            elif kind == '8':
                tname = 'adStatus'  # administratal status
            elif kind == '13':
                tname = 'in_discards'  # counter32
            elif kind == '14':
                tname = 'in_error'  # counter32
            elif kind == '19':
                tname = 'out_discards'  # counter32
            elif kind == '20':
                tname = 'out_error'  # counter32
            else:
                continue
            results[name][tname] = result_all[oid]

        # IF-MIB::ifMIB_table = '1.3.6.1.2.1.31'
        base = '1.3.6.1.2.1.31.1.1.1'
        base_idx = '1.3.6.1.2.1.2.2.1.1.'
        base_name = base + '.1.'
        # results = {}
        idx2name = {}
        result_all = self._get_walk(base)
        for oid in sorted(result_all.keys()):
            if base_idx in oid:
                val = str(result_all[oid])
                idx2name[val] = None
            elif base_name in oid:
                idx = oid.split('.')[-1]
                val = result_all[oid]
                idx2name[idx] = val
        for oid in sorted(result_all.keys()):
            if oid in base_idx or oid in base_name:
                continue
            idx = oid.split('.')[-1]
            name = str(idx2name[idx])
            results[name] = results.get(name, {})
            tname = oid
            kind = oid.split('.')[-2]
            if kind == '6':
                tname = 'in_octets'  # counter64
            elif kind == '10':
                tname = 'out_octets'  # counter64
            elif kind == '15':
                tname = 'speed'  # gauge32
            else:
                continue
            results[name][tname] = result_all[oid]
        return results

    def postgres_backup(self):
        # val = self._get_mib('public', 'NET-SNMP-EXTEND-MIB', 'check_postgres_backup')
        # res = self.parse(OctetString, str, val)
        # Workaround for a too large check response from the script(too large udp packets are blocked by the fw, so we use tcp)
        cmd = \
            '/usr/bin/snmpget -v2c -c public -t {timeout} tcp:{host} \'NET-SNMP-EXTEND-MIB::nsExtendOutputFull."check_postgres_backup"\''.format(timeout=self.timeout,
                host=self.host)
        try:
            output = subprocess32.check_output(cmd, shell=True, timeout=self.timeout, stderr=subprocess32.STDOUT)
        except subprocess32.CalledProcessError, e:
            database_backup_size, warnings, check_duration = {}, [str(e.output)], 0
        else:
            res = output.partition(': ')[2]
            s = res.split('|')
            warning = s[0]
            perfdata = s[-1]
            database_backup_size = {}
            check_duration = 0
            for pd in perfdata.split(' '):
                k, _, v = pd.partition('=')
                v = v.split(';')[0]
                if k == 'time':
                    check_duration = v
                else:
                    database_backup_size[k] = v
            warnings = warning.split(';')
        return {'database_backup_size': database_backup_size, 'warnings': warnings, 'check_duration': check_duration}

    def disk_pgxlog(self):
        result = {}
        output = str(self._get_mib('public', 'NET-SNMP-EXTEND-MIB', 'disk_pgxlog'))
        output = [i for i in re.split('\s{1,}|\t|\n', output) if i.isdigit() or i.startswith('/')]
        output = [output[i:i + 7] for i in range(0, len(output), 7)]
        output = [[
            i[1],
            int(int(i[0]) / 1024),
            i[2],
            int(i[3]),
            int(i[4]),
            int(i[5]),
            i[6],
        ] for i in output]
        for i in output:
            name = str(i[0])
            result[name] = {}
            for index in range(len(i)):
                val = i[index]
                if val == i[1]:
                    tname = 'du_dir'
                elif val == i[2]:
                    tname = 'filesystem'
                elif val == i[3]:
                    tname = 'totalspace'
                elif val == i[4]:
                    tname = 'used_space'
                elif val == i[5]:
                    tname = 'available_space'
                elif val == i[6]:
                    tname = 'mounted_on'
                else:
                    continue
                result[name][tname] = i[index]

        return result

    def conntrackstats(self):
        output = str(self._get_mib('public', 'NET-SNMP-EXTEND-MIB', 'conntrackstats'))
        return dict((a.split('=')[0].strip(), int(a.split('=')[1])) for a in output.split('|'))

    #
    # snmp functions
    #

    def get_list(self, prefix='get', convert=int, *oids):
        h = {}
        for idx, oid in enumerate(oids):
            k = '{0}{1}'.format(prefix, idx)
            h[k] = self.get(oid, k, convert)
        return h

    def get(self, oid, name, convert):
        '''Example:
        get('1.3.6.1.4.1.2021.4.6.0', 'foo', int) -> {'foo': 42}
        '''

        val = self._get_cmd(oid)
        result = convert(val)
        return {name: result}

    def _get_mib(self, community, prefix, extend, mib_result='nsExtendOutputFull', path=None):
        '''Parameter: mib_result can be 'nsExtendOutputFull', 'nsExtendResult', ..
        Example:
        _get_mib("public", "NET-SNMP-EXTEND-MIB", "check_postgres_backup")
        is the same like:
        % snmpget -v2c -c public z-storage02 'NET-SNMP-EXTEND-MIB::nsExtendOutputFull."check_postgres_backup"'
        '''

        mib = None
        if path:
            mib = cmdgen.MibVariable(prefix, mib_result, extend).addMibSource(path)
        else:
            mib = cmdgen.MibVariable(prefix, mib_result, extend)
        real_fun = getattr(self.generator, 'getCmd')  # SNMPGET
        res = errorIndication, errorStatus, errorIndex, varBinds = real_fun(self.comm_data, self.transport, mib)

        if not errorIndication is None or errorStatus is True:
            msg = 'Error: %s %s %s %s' % res
            raise SnmpError(msg)
        else:
            _, val = varBinds[0]
            return val

    def _get_mib_bulkwalk(self, community, prefix, table, path=None):
        '''
        % snmpbulkwalk -v2c -c public myhost123 TCP-MIB::tcpConnTable
        '''

        mib = None
        if path:
            mib = cmdgen.MibVariable(prefix, table).addMibSource(path)
        else:
            mib = cmdgen.MibVariable(prefix, table)
        real_fun = getattr(self.generator, 'bulkCmd')  # SNMPBULKWALK
        res = errorIndication, errorStatus, errorIndex, varBinds = real_fun(self.comm_data, self.transport, 0, 50, mib,
                max_rows=100, ignore_non_increasing_oid=True)

        if not errorIndication is None or errorStatus is True:
            msg = 'Error: %s %s %s %s' % res
            raise SnmpError(msg)
        else:
            res = {}
            for items in varBinds:
                oid = str(items[0][0])
                val = items[0][1]
                if isinstance(val, Counter64) or isinstance(val, Counter32) or isinstance(val, Integer):
                    res[oid] = int(val)
                else:
                    res[oid] = str(val)
            return res

    def _get_walk(self, oid):
        '''Returns a dictionary of oid -> value'''

        real_fun = getattr(self.generator, 'nextCmd')  # SNMPWALK
        res = errorIndication, errorStatus, errorIndex, varBinds = real_fun(self.comm_data, self.transport, oid)
        if not errorIndication is None or errorStatus is True:
            msg = 'Error: %s %s %s %s' % res
            raise SnmpError(msg)
        else:
            res = {}
            for items in varBinds:
                oid = str(items[0][0])
                val = items[0][1]
                if isinstance(val, Counter64) or isinstance(val, Counter32) or isinstance(val, Integer):
                    res[oid] = int(val)
                else:
                    res[oid] = str(val)
            return res

    def _get_cmd(self, oid):
        real_fun = getattr(self.generator, 'getCmd')  # SNMPGET
        res = errorIndication, errorStatus, errorIndex, varBinds = real_fun(self.comm_data, self.transport, oid)

        if not errorIndication is None or errorStatus is True:
            msg = 'Error: %s %s %s %s' % res
            raise SnmpError(msg)
        else:
            _, val = varBinds[0]
            return val

    # Convert SNMP data to python data
    def parse(self, clazz, convert, val):
        '''Example: self.parse(Integer, int, Integer(11040956))'''

        if not val:
            return None
        if isinstance(val, clazz):
            return convert(val)
        raise SnmpError('Could not convert [{}] with {} into {}'.format(val, convert, clazz))





#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Query Exasol
"""

import tempfile
import subprocess
import os

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial


class ExaplusFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(ExaplusFactory, self).__init__()
        # fields from config
        self._exacrm_cluster = None
        self._exacrm_user = None
        self._exacrm_pass = None

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        self._exacrm_cluster = conf['exacrm_cluster']
        self._exacrm_user = conf['exacrm_user']
        self._exacrm_pass = conf['exacrm_pass']

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(ExaplusWrapper, cluster=self._exacrm_cluster, password=self._exacrm_pass,
                          user=self._exacrm_user)


class ExaplusWrapper(object):

    def __init__(self, cluster, user='ZALANDO_NAGIOS', password='', schema=False):
        self._err = None
        self._out = None
        self.user = user
        self.__password = password
        self.cluster = cluster
        self.schema = schema
        self.java_opts = ['-Djava.net.preferIPv4Stack=true', '-Djava.awt.headless=true', '-Xmx512m', '-Xms128m']
        self.exaplus_opts = [
            '-recoverConnection',
            'OFF',
            '-retry',
            '0',
            '-lang',
            'EN',
            '-q',
            '-x',
            '-Q',
            '10',
            '-pipe',
        ]
        self.jar_file = '/server/exasol/exaplus/current/exaplus.jar'

    def query(self, query):
        fd, name = tempfile.mkstemp(suffix='.sql', text=True)
        try:
            fh = os.fdopen(fd, 'w')
            fh.write('%s\n' % query)
            fh.flush()

            cmd = ['/usr/bin/java']
            cmd.extend(self.java_opts)
            cmd.extend(['-jar', self.jar_file])
            cmd.extend(['-c', self.cluster])
            cmd.extend(['-u', self.user])
            cmd.extend(['-p', self.__password])
            cmd.extend(self.exaplus_opts)
            if self.schema:
                cmd.extend(['-s', self.schema])
            cmd.extend(['-f', name])
            # print "EXAPLUS="+" ".join(cmd)
            sub = subprocess.Popen(cmd, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
            d_out, d_err = sub.communicate()
            self._out = d_out
            self._err = d_err
        finally:
            os.unlink(name)
        return self

    def result(self):
        return self._out.split('\n'), self._err.split('\n')


if __name__ == '__main__':
    import sys
    exaplus = ExaplusWrapper(sys.argv[1], sys.argv[2], sys.argv[3])
    print exaplus.query('''select table_name, case when hours_between(systimestamp,last_export_success_time) < 24 then 'EXPORTED YESTERDAY (within 24 HOURS) - OK' else 'NOT EXPORTED within LAST 24 HOURS' end status, last_export_success_time LAST_EXPORT_TIME from STG.TARGET_LOADING_STATUS where table_name not in ('D_SHOP','F_CUSTOMER_SALES','F_PRODUCT_AVAILABILITY','F_UMS_CAMPAIGN_RESPONSE') order by 1 ;'''
                        ).result()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Our thin layer on top of yapsy_plugin to load external code
Folders to be explored are taken from the environment variable ZMON_PLUGINS

TODO: Add examples and point to the tests
"""

import logging
import os
import sys

from yapsy.PluginManager import PluginManagerSingleton
import pkg_resources

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin

logger = logging.getLogger(__name__)


# Some constants to define global behaviour

PLUGIN_INFO_EXT = 'worker_plugin'

PLUGIN_BUILTINS = ('zmon_worker_monitor.builtins.plugins',)

PLUGIN_ENV_VAR = 'ZMON_PLUGINS'

PLUGIN_CATEGORIES_FILTER = {
    'Function': IFunctionFactoryPlugin,
}

GLOBAL_CONFIG_PREFIX = 'plugin.{plugin_name}.'


class PluginError(Exception):
    pass


class PluginRecoverableError(PluginError):
    pass


class PluginFatalError(PluginError):
    pass


def _builtins_paths(subpackages, raise_errors=True):
    if not subpackages:
        return []
    folders = []

    for subpkg in (subpackages if isinstance(subpackages, (tuple, list)) else [subpackages]):
        try:
            parts = subpkg.split('.')
            path = pkg_resources.resource_filename('.'.join(parts[:-1]), parts[-1])
            if not os.path.isdir(path):
                raise Exception('path is not a directory: {}'.format(path))
        except Exception:
            logger.exception('erroneous plugins package: %s. Exception: ', subpkg)
            if raise_errors:
                _, ev, tb = sys.exc_info()
                raise PluginFatalError('Builtins plugins error in {}. Reason: {}'.format(subpkg, ev)), None, tb
        else:
            folders.append(path)

    return folders


def _env_dirs(env_var, raise_errors=True):
    env_value = os.environ.get(env_var)
    if not env_value:
        return []

    folders = []
    for d in env_value.split(os.pathsep):
        if not os.path.isdir(d):
            logger.warn('Wrong path %s in env variable %s', d, env_var)
            if raise_errors:
                raise PluginFatalError('Env plugins error in path: {}, from env_var: {}'.format(d, env_var))
            continue
        folders.append(d)
    return folders


def _filter_additional_dirs(path_list, raise_errors=True):
    if not path_list:
        return []
    folders = []
    for path in path_list:
        if os.path.isdir(path):
            folders.append(path)
        elif raise_errors:
            raise PluginFatalError('Additional dirs contains erroneous path: {}'.format(path))
    return folders


_initialized = {}


def init_plugin_manager(category_filter=None, info_ext=PLUGIN_INFO_EXT, builtins_pkg=PLUGIN_BUILTINS,
                        env_var=PLUGIN_ENV_VAR):
    """
    Initialize the plugin manager and set some behaviour options
    :param category_filter:
    :param info_ext:
    :param builtins_pkg:
    :param env_var:
    :return:
    """
    global _initialized

    # default category_filter is PLUGIN_CATEGORIES_FILTER (dict)
    category_filter = PLUGIN_CATEGORIES_FILTER if category_filter is None else category_filter

    logger.info('init plugin manager')

    manager = PluginManagerSingleton.get()

    manager.setCategoriesFilter(category_filter)
    manager.setPluginInfoExtension(info_ext)

    # save parameters used to initialize the module
    _initialized = dict(category_filter=category_filter, info_ext=info_ext, builtins_pkg=builtins_pkg, env_var=env_var)


def get_plugin_manager():
    """
    Get the plugin manager object (singleton)
    """
    return PluginManagerSingleton.get()


_collected = False


def collect_plugins(load_builtins=True, load_env=True, additional_dirs=None, global_config=None, raise_errors=True):
    """
    Collect plugins from folders in environment var and additional_dir param.

    :param plugin_env_var: environment variable containing a list of paths (shell $PATH style)
    :param additional_dirs: additional locations to search plugins in
    :return:
    """
    global _collected

    if not _initialized:
        raise PluginFatalError('You must invoke init_plugin_manager() before collect_plugins()!')
    if _collected:
        raise PluginFatalError('Plugins should be collected only once!')

    try:
        # load the plugins

        global_config = {} if global_config is None else global_config

        builtins = _builtins_paths(_initialized.get('builtins_pkg')) if load_builtins else []

        paths_env = _env_dirs(_initialized.get('env_var')) if load_env else []

        path_list = paths_env + _filter_additional_dirs(additional_dirs)

        # not necessary and may cause module name clashes... remove?
        # for entry in path_list:
        #    if entry not in sys.path:
        #        sys.path.append(entry)  # so the plugins can relatively import their submodules

        # check plugin dependencies declared in {plugin_dir}/requirements.txt are installed
        for path in path_list:
            miss_deps = _check_dependencies(path)
            if miss_deps:
                logger.error('Dependencies missing for plugin %s: %s', path, ','.join(miss_deps))
                if raise_errors:
                    raise PluginFatalError('Dependencies missing for plugin {}: {}'.format(path, ','.join(miss_deps)))

        manager = get_plugin_manager()
        manager.setPluginPlaces(builtins + path_list)

        # explore the provided locations and identify plugin candidates
        manager.locatePlugins()

        # save list of all plugin candidates: [(info file path, python file path, plugin info instance), ...]
        candidates = manager.getPluginCandidates()

        logger.debug('Recognized plugin candidates: %s', candidates)

        # trigger the loading of all plugin python modules
        manager.loadPlugins()

        all_plugins = manager.getAllPlugins()

        if len(all_plugins) != len(candidates):
            plugin_paths = map(_path_source_to_plugin, [p.path for p in all_plugins])

            dropped = [c for c in candidates if c[0] not in plugin_paths]
            logger.error('These plugin candidates have errors: %s', dropped)
            if raise_errors:
                raise PluginFatalError('Plugin candidates have errors: {}'.format(dropped))

        # configure and activate plugins

        for plugin in all_plugins:

            config_prefix = GLOBAL_CONFIG_PREFIX.format(plugin_name=plugin.name)

            conf_global = {}
            try:
                conf_global = {str(c)[len(config_prefix):]: v for c, v in global_config.iteritems() if str(c).startswith(config_prefix)}
                logger.debug('Plugin %s received global conf keys: %s', plugin.name, conf_global.keys())
            except Exception:
                logger.exception('Failed to parse global configuration. Reason: ')
                if raise_errors:
                    raise

            conf = {}
            try:
                if plugin.details.has_section('Configuration'):
                    conf = {c: v for c, v in plugin.details.items('Configuration')}  # plugin.plugin_info.detail has the safeconfig object
                logger.debug('Plugin %s received local conf keys: %s', plugin.name, conf.keys())
            except Exception:
                logger.exception('Failed to load local configuration from plugin: %s. Reason: ', plugin.name)
                if raise_errors:
                    raise

            # for security reasons our global config take precedence over the local config
            conf.update(conf_global)

            try:
                plugin.plugin_object.configure(conf)
            except Exception:
                logger.exception('Failed configuration of plugin: %s. Reason: ', plugin.name)
                if raise_errors:
                    raise
                plugin.plugin_object.deactivate()
                continue

            plugin.plugin_object.activate()

        _collected = True

    except PluginFatalError:
        raise

    except Exception:
        logger.exception('Unexpected error during plugin collection: ')
        if raise_errors:
            _, ev, tb = sys.exc_info()
            raise PluginFatalError("Error while loading plugins. Reason: {}".format(ev)), None, tb


def get_plugins_of_category(category, active=True, raise_errors=True):
    """
    Get plugins (plugin_info) of a given category
    """
    try:
        plugins = get_plugin_manager().getPluginsOfCategory(category)
    except KeyError:
        if raise_errors:
            raise PluginRecoverableError('Category {} not known to the plugin system'.format(category))
        return []
    if plugins and isinstance(active, bool):
        plugins = [p for p in plugins if p.is_activated == active]
    return plugins


def get_plugin_objs_of_category(category, active=True, raise_errors=True):
    """
    Get plugin objects of a given category
    """
    return [p.plugin_object for p in get_plugins_of_category(category, active, raise_errors)]


def get_plugin_by_name(name, category, not_found_is_error=True):
    """
    Get a plugin by name and category
    """
    plugin = get_plugin_manager().getPluginByName(name, category)

    if not plugin and not_found_is_error:
        raise PluginRecoverableError('Plugin by name {} not found under category {}'.format(name, category))

    return plugin


def get_plugin_obj_by_name(name, category, not_found_is_error=True):
    """
    Get a plugin object by name and category
    """
    plugin = get_plugin_by_name(name, category, not_found_is_error)
    return None if plugin is None else plugin.plugin_object


def get_all_plugins():
    """
    Get list of all loaded plugins
    """
    return get_plugin_manager().getAllPlugins()


def get_all_plugin_names():
    """
    Get list of names of all loaded plugins
    """
    plugins = get_all_plugins()
    if not plugins:
        return []
    return [p.name for p in plugins]


def get_all_categories():
    """
    Return list of categories provided in the category_filter
    """
    manager = get_plugin_manager()
    return manager.getCategories()


def get_loaded_plugins_categories():
    """
    Return list of categories that have one or more discovered plugins
    """
    return list(set([p.category for p in get_all_plugins()]))


def _check_dependencies(path):

    req_path = path + os.sep + 'requirements.txt'
    if not os.path.isfile(req_path):
        logger.debug('%s has no requirements.txt file' % path)
        return None

    missing_pkg = []
    with open(req_path) as f:
        for line in f:
            stripped = line.strip()
            if stripped and not stripped.startswith('#'):
                try:
                    pkg_resources.get_distribution(stripped)
                except Exception as _:
                    missing_pkg.append(stripped)
    return missing_pkg


def _path_source_to_plugin(source_path):
    source_no_py = source_path[:-3] if source_path.lower().endswith('.py') else source_path
    plugin_path = source_no_py + '.' + _initialized['info_ext']
    return plugin_path



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import ast
from inspect import isclass
import __future__

from collections import Callable, Counter
import socket
from zmon_worker_monitor.zmon_worker.encoder import JsonDataEncoder
from stashacc import StashAccessor
from zmon_worker_monitor.zmon_worker.common.utils import async_memory_cache, with_retries
from zmon_worker_monitor.zmon_worker.errors import *

import zmon_worker_monitor.eventloghttp as eventlog
import functools
import itertools
import json
import logging
import random
from zmon_worker_monitor.redis_context_manager import RedisConnHandler
import time
import re
import requests
import sys
import setproctitle
from datetime import timedelta, datetime
import urllib
import pytz
import threading
import Queue
from collections import defaultdict

from bisect import bisect_left
from zmon_worker_monitor.zmon_worker.common.time_ import parse_timedelta

from zmon_worker_monitor.zmon_worker.notifications.hipchat import NotifyHipchat
from zmon_worker_monitor.zmon_worker.notifications.slack import NotifySlack
from zmon_worker_monitor.zmon_worker.notifications.mail import Mail
from zmon_worker_monitor.zmon_worker.notifications.sms import Sms
from zmon_worker_monitor.zmon_worker.notifications.hubot import Hubot
from zmon_worker_monitor.zmon_worker.notifications.notification import BaseNotification

from operator import itemgetter
from timeperiod import in_period, InvalidFormat

import functional
from zmon_worker_monitor.zmon_worker.common import mathfun

from zmon_worker_monitor import plugin_manager


logger = logging.getLogger(__name__)


# interval in seconds for sending metrics to graphite
METRICS_INTERVAL = 15
STASH_CACHE_EXPIRATION_TIME = 3600

DEFAULT_CHECK_RESULTS_HISTORY_LENGTH = 20

TRIAL_RUN_RESULT_EXPIRY_TIME = 300

# we allow specifying condition expressions without using the "value" variable
# the following pattern is used to check if "value" has to be prepended to the condition
SIMPLE_CONDITION_PATTERN = re.compile(r'^[<>!=\[]|i[ns] ')
GRAPHITE_REPLACE_KEYCHARS = re.compile(r'[./\s]')

# round to microseconds
ROUND_SECONDS_DIGITS = 6
JMX_CONFIG_FILE = 'jmxremote.password'
KAIROS_ID_FORBIDDEN_RE = re.compile(r'[^a-zA-Z0-9\-_\.]')

HOST_GROUP_PREFIX = re.compile(r'^([a-z]+)')
INSTANCE_PORT_SUFFIX = re.compile(r':([0-9]+)$')

EVENTS = {
    'ALERT_STARTED': eventlog.Event(0x34001, ['checkId', 'alertId', 'value']),
    'ALERT_ENDED': eventlog.Event(0x34002, ['checkId', 'alertId', 'value']),
    'ALERT_ENTITY_STARTED': eventlog.Event(0x34003, ['checkId', 'alertId', 'value', 'entity']),
    'ALERT_ENTITY_ENDED': eventlog.Event(0x34004, ['checkId', 'alertId', 'value', 'entity']),
    'DOWNTIME_STARTED': eventlog.Event(0x34005, [
        'alertId',
        'entity',
        'startTime',
        'endTime',
        'userName',
        'comment',
    ]),
    'DOWNTIME_ENDED': eventlog.Event(0x34006, [
        'alertId',
        'entity',
        'startTime',
        'endTime',
        'userName',
        'comment',
    ]),
    'SMS_SENT': eventlog.Event(0x34007, ['alertId', 'entity', 'phoneNumber', 'httpStatus']),
    'ACCESS_DENIED': eventlog.Event(0x34008, ['userName', 'entity']),
}

eventlog.register_all(EVENTS)

Sms.register_eventlog_events(EVENTS)

get_value = itemgetter('value')


class ProtectedPartial(object):

    '''
    Provides functools.partial functionality with one additional feature: if keyword arguments contain '__protected'
    key with list of arguments as value, the appropriate values will not be overwritten when calling the partial. This
    way we can prevent user from overwriting internal zmon parameters in check command. The protected key uses double
    underscore to prevent overwriting it, we reject all commands containing double underscores.
    '''

    def __init__(self, func, *args, **kwargs):
        self.__func = func
        self.__partial_args = args
        self.__partial_kwargs = kwargs
        self.__protected = frozenset(kwargs.get('__protected', []))
        self.__partial_kwargs.pop('__protected', None)

    def __call__(self, *args, **kwargs):
        new_kwargs = self.__partial_kwargs.copy()
        new_kwargs.update((k, v) for (k, v) in kwargs.iteritems() if k not in self.__protected)
        return self.__func(*self.__partial_args + args, **new_kwargs)



def propartial(func, *args, **kwargs):
    '''
    >>> propartial(int, base=2)('100')
    4
    >>> propartial(int, base=2)('100', base=16)
    256
    >>> propartial(int, base=2, __protected=['base'])('100', base=16)
    4
    '''

    return ProtectedPartial(func, *args, **kwargs)


normalize_kairos_id = propartial(KAIROS_ID_FORBIDDEN_RE.sub, '_')

orig_process_title = None


def setp(check_id, entity, msg):
    global orig_process_title
    if not orig_process_title:
        try:
            orig_process_title = setproctitle.getproctitle().split(' ')[2].split(':')[0].split('.')[0]
        except:
            orig_process_title = 'p34XX'

    setproctitle.setproctitle('zmon-worker.{} check {} on {} {} {}'.format(orig_process_title, check_id, entity, msg,
                              datetime.now().strftime('%H:%M:%S.%f')))


def flatten(structure, key='', path='', flattened=None):
    path = str(path)
    key = str(key)

    if flattened is None:
        flattened = {}
    if type(structure) not in (dict, list):
        flattened[((path + '.' if path else '')) + key] = structure
    elif isinstance(structure, list):
        pass
    else:
        for new_key, value in structure.items():
            flatten(value, new_key, '.'.join(filter(None, [path, key])), flattened)
    return flattened

def timed(f):
    '''Decorator to "time" a function execution. Wraps the function's result in a new dict.
    >>> timed(lambda: 1)()['value']
    1
    '''

    def wrapper(*args, **kwargs):
        start = time.time()
        res = f(*args, **kwargs)
        delta = time.time() - start
        # round and use short keys as we will serialize the whole stuff as JSON
        return {'value': res, 'ts': round(start, ROUND_SECONDS_DIGITS), 'td': round(delta, ROUND_SECONDS_DIGITS)}

    return wrapper


def _get_entity_url(entity):
    '''
    >>> _get_entity_url({})

    >>> _get_entity_url({'url': 'fesn01:39820'})
    'http://fesn01:39820'

    >>> _get_entity_url({'host': 'fesn01'})
    'http://fesn01'
    '''

    if 'url' in entity:
        if entity['url'].startswith('http://') or entity['url'].startswith('https://'):
            return entity['url']

        return 'http://' + entity['url']

    if 'host' in entity:
        return 'http://' + entity['host']

    return None


def _get_jmx_port(entity):
    '''
    >>> _get_jmx_port({'instance': '9620'})
    49620
    '''

    if 'instance' in entity:
        return int('4' + entity['instance'])
    return None


def _get_shards(entity):
    '''
    >>> _get_shards({'shards': {'shard1': 'host1/db1'}})
    {'shard1': 'host1/db1'}

    >>> _get_shards({'service_name': 'db'})
    {'db': 'db/postgres'}

    >>> _get_shards({'service_name': 'db', 'port': 1234})
    {'db': 'db:1234/postgres'}

    >>> _get_shards({'service_name': 'db:1234', 'port': 1234})
    {'db:1234': 'db:1234/postgres'}

    >>> _get_shards({'service_name': 'db:1234'})
    {'db:1234': 'db:1234/postgres'}

    >>> _get_shards({'service_name': 'db-1234', 'port': 1234})
    {'db-1234': 'db-1234:1234/postgres'}

    >>> _get_shards({'project': 'shop'})
    '''

    if 'shards' in entity:
        return entity['shards']
    if 'service_name' in entity:
        return {entity['service_name']: ('{service_name}:{port}/postgres'.format(**entity) if 'port' in entity
                and not entity['service_name'].endswith(':{}'.format(entity['port'
                ])) else '{}/postgres'.format(entity['service_name']))}
    return None


def entity_values(con, check_id, alert_id,count=1):
    return map(get_value, entity_results(con, check_id, alert_id, count))


def entity_results(con, check_id, alert_id, count=1):
    all_entities = con.hkeys('zmon:alerts:{}:entities'.format(alert_id))
    all_results = []
    for entity_id in all_entities:
        results = get_results(con, check_id, entity_id, count)
        all_results.extend(results)
    return all_results


def capture(value=None, captures=None, **kwargs):
    '''
    >>> capture(1, {})
    1

    >>> captures={}; capture(1, captures); captures
    1
    {'capture_1': 1}

    >>> captures={'capture_1': 1}; capture(2, captures); sorted(captures.items())
    2
    [('capture_1', 1), ('capture_2', 2)]

    >>> captures={}; capture(captures=captures, mykey=1); captures
    1
    {'mykey': 1}

    >>> p = functools.partial(capture, captures={}); p(1); p(a=1)
    1
    1
    '''

    if kwargs:
        if len(kwargs) > 1:
            raise ValueError('Only one named capture supported')
        key, value = kwargs.items()[0]
    else:
        i = 1
        while True:
            key = 'capture_{}'.format(i)
            if key not in captures:
                break
            i += 1
    captures[key] = value
    return value


def _parse_alert_parameter_value(data):
    '''
    >>> _parse_alert_parameter_value({'value': 10})
    10
    >>> _parse_alert_parameter_value({'value': '2014-07-03T22:00:00.000Z', 'comment': "desc", "type": "date"})
    datetime.date(2014, 7, 3)

    '''

    allowed_types = {
        'int': int,
        'float': float,
        'str': str,
        'bool': bool,
        'datetime': lambda json_date: datetime.strptime(json_date, '%Y-%m-%dT%H:%M:%S.%fZ'),
        'date': lambda json_date: datetime.strptime(json_date, '%Y-%m-%dT%H:%M:%S.%fZ').date(),
    }
    value = data['value']
    type_name = data.get('type')
    if type_name:
        try:
            value = allowed_types[type_name](value)
        except Exception:
            raise Exception('Attempted wrong type cast <{}> in alert parameters'.format(type_name))
    return value


def _inject_alert_parameters(alert_parameters, ctx):
    '''
    Inject alert parameters into the execution context dict (ctx)
    '''

    params_name = 'params'
    params = {}

    if alert_parameters:
        for apname, apdata in alert_parameters.items():
            if apname in ctx:
                raise Exception('Parameter name: %s clashes in context', apname)
            value = _parse_alert_parameter_value(apdata)
            params[apname] = value
            ctx[apname] = value  # inject parameter into context

        # inject the whole parameters map so that user can iterate over them in the alert condition
        if params_name not in ctx:
            ctx[params_name] = params

def alert_series(f, n, con, check_id, entity_id):
    """ evaluate given function on the last n check results and return true if the "alert" function f returns true for all values"""

    vs = get_results(con, check_id, entity_id, n)
    active_count = 0
    exception_count = 0

    for v in vs:
        # counting exceptions thrown during eval as alert being active for that interval
        try:
            v = v["value"]
            r = 1 if f(v) else 0
            x =0
        except:
            r = 1
            x = 1

        active_count += r
        exception_count += x

    if exception_count == n:
        raise Exception("All alert evaluations failed!")

    # activating alert if not enough value found (this should only affect starting period)
    return n == active_count or len(vs)<n

def build_condition_context(con, check_id, alert_id, entity, captures, alert_parameters):
    '''
    >>> plugin_manager.collect_plugins(); 'timeseries_median' in build_condition_context(None, 1, 1, {'id': '1'}, {}, {})
    True
    >>> 'timeseries_percentile' in build_condition_context(None, 1, 1, {'id': '1'}, {}, {})
    True
    '''

    history_factory = plugin_manager.get_plugin_obj_by_name('history', 'Function')

    ctx = build_default_context()
    ctx['capture'] = functools.partial(capture, captures=captures)
    ctx['entity_results'] = functools.partial(entity_results, con=con, check_id=check_id, alert_id=alert_id)
    ctx['entity_values'] = functools.partial(entity_values, con=con, check_id=check_id, alert_id=alert_id)
    ctx['entity'] = dict(entity)
    ctx['history'] = history_factory.create({ 'check_id': check_id, 'entity_id_for_kairos': normalize_kairos_id(entity['id']) })
    ctx['value_series'] = functools.partial(get_results_user, con=con, check_id=check_id, entity_id=entity['id'])
    ctx['alert_series'] = functools.partial(alert_series, con=con, check_id=check_id, entity_id=entity['id'])

    _inject_alert_parameters(alert_parameters, ctx)

    for f in (
        mathfun.avg,
        mathfun.delta,
        mathfun.median,
        mathfun.percentile,
        mathfun.first,
        mathfun._min,
        mathfun._max,
        sum,
    ):
        name = f.__name__
        if name.startswith('_'):
            name = name[1:]
        ctx['timeseries_' + name] = functools.partial(_apply_aggregate_function_for_time, con=con, func=f, check_id=check_id,
                                            entity_id=entity['id'], captures=captures)
    return ctx


def _time_slice(time_spec, results):
    '''
    >>> _time_slice('1s', [])
    []

    >>> _time_slice('1s', [{'ts': 0, 'value': 0}, {'ts': 1, 'value': 10}])
    [{'ts': 0, 'value': 0}, {'ts': 1, 'value': 10}]

    >>> _time_slice('2s', [{'ts': 123.6, 'value': 10}, {'ts': 123, 'value': 0}, {'ts': 121, 'value': -10}])
    [{'ts': 123, 'value': 0}, {'ts': 123.6, 'value': 10}]
    '''

    if len(results) < 2:
        # not enough values to calculate anything
        return results
    get_ts = itemgetter('ts')
    results.sort(key=get_ts)
    keys = map(get_ts, results)
    td = parse_timedelta(time_spec)
    last = results[-1]
    needle = last['ts'] - td.total_seconds()
    idx = bisect_left(keys, needle)
    if idx == len(results):
        # timerange exceeds range of results
        return results
    return results[idx:]


def _get_results_for_time(con, check_id, entity_id, time_spec):
    results = get_results(con, check_id, entity_id, DEFAULT_CHECK_RESULTS_HISTORY_LENGTH)
    return _time_slice(time_spec, results)


def _apply_aggregate_function_for_time(
    time_spec,
    con,
    func,
    check_id,
    entity_id,
    captures,
    key=functional.id,
    **args
):

    results = _get_results_for_time(con, check_id, entity_id, time_spec)
    ret = mathfun.apply_aggregate_function(results, func, key=functional.compose(key, get_value), **args)
    # put function result in our capture dict for debugging
    # e.g. captures["delta(5m)"] = 13.5
    captures['{}({})'.format(func.__name__, time_spec)] = ret
    return ret


def _build_notify_context(alert):
    return {
            'send_mail': functools.partial(Mail.send, alert),
            'send_email': functools.partial(Mail.send, alert),
            'send_sms': functools.partial(Sms.send, alert),
            'notify_hubot': functools.partial(Hubot.notify, alert),
            'send_hipchat': functools.partial(NotifyHipchat.send, alert),
            'send_slack': functools.partial(NotifySlack.send, alert)
           }


def _prepare_condition(condition):
    '''function to prepend "value" to condition if necessary

    >>> _prepare_condition('>0')
    'value >0'

    >>> _prepare_condition('["a"]>0')
    'value ["a"]>0'

    >>> _prepare_condition('in (1, 2, 3)')
    'value in (1, 2, 3)'

    >>> _prepare_condition('value>0')
    'value>0'

    >>> _prepare_condition('a in (1, 2, 3)')
    'a in (1, 2, 3)'
    '''

    if SIMPLE_CONDITION_PATTERN.match(condition):
        # short condition format, e.g. ">=3"
        return 'value {}'.format(condition)
    else:
        # condition is more complex, e.g. "value > 3 and value < 10"
        return condition


class PeriodicBufferedAction(object):

    def __init__(self, action, action_name=None, retries=5, t_wait=10, t_random_fraction=0.5):

        self._stop = True
        self.action = action
        self.action_name = action_name if action_name else (action.func_name if hasattr(action, 'func_name') else
                                                            (action.__name__ if hasattr(action, '__name__') else None))
        self.retries = retries
        self.t_wait = t_wait
        self.t_rand_fraction = t_random_fraction

        self._queue = Queue.Queue()
        self._thread = threading.Thread(target=self._loop)
        self._thread.daemon = True

    def start(self):
        self._stop = False
        self._thread.start()

    def stop(self):
        self._stop = True

    def is_active(self):
        return not self._stop

    def get_time_randomized(self):
        return self.t_wait * (1 + random.uniform(-self.t_rand_fraction, self.t_rand_fraction))

    def enqueue(self, data, count=0):
        elem = {
            'data': data,
            'count': count,
            # 'time': time.time()
        }
        try:
            self._queue.put_nowait(elem)
        except Queue.Full:
            logger.exception('Fatal Error: is worker out of memory? Details: ')

    def _collect_from_queue(self):
        elem_list = []
        empty = False

        while not empty and not self._stop:
            try:
                elem_list.append(self._queue.get_nowait())
            except Queue.Empty:
                empty = True
        return elem_list

    def _loop(self):
        t_last = time.time()
        t_wait_last = self.get_time_randomized()

        while not self._stop:
            if time.time() - t_last >= t_wait_last:
                elem_list = self._collect_from_queue()
                try:
                    if elem_list:
                        self.action([e['data'] for e in elem_list])
                except Exception as e:
                    logger.error('Error executing action %s: %s', self.action_name, e)
                    for elem in elem_list:
                        if elem['count'] < self.retries:
                            self.enqueue(elem['data'], count=elem['count']+1)
                        else:
                            logger.error('Error: Maximum retries reached for action %s. Dropping data: %s ',
                                         self.action_name, elem['data'])
                finally:
                    t_last = time.time()
                    t_wait_last = self.get_time_randomized()
            else:
                time.sleep(0.2)  # so loop is responsive to stop commands


def _log_event(event_name, alert, result, entity=None):
    params = {'checkId': alert['check_id'], 'alertId': alert['id'], 'value': result['value']}

    if entity:
        params['entity'] = entity

    eventlog.log(EVENTS[event_name].id, **params)


def _convert_captures(worker_name, alert_id, entity_id, timestamp, captures):
    '''
    >>> _convert_captures('p0.h', 1, 'e1', 1, {'c0': 'error'})
    []
    >>> _convert_captures('p0.h', 1, 'e1', 1, {'c1': '23.4'})
    [('p0_h.alerts.1.e1.captures.c1', 23.4, 1)]
    >>> _convert_captures('p0.h', 1, 'e1', 1, {'c2': 12})
    [('p0_h.alerts.1.e1.captures.c2', 12.0, 1)]
    >>> _convert_captures('p0.h', 1, 'e1', 1, {'c3': {'c31': '42'}})
    [('p0_h.alerts.1.e1.captures.c3.c31', 42.0, 1)]
    >>> _convert_captures('p0.h', 1, 'e1', 1, {'c4': {'c41': 'error'}})
    []
    >>> _convert_captures('p0.h', 1, 'e .1/2', 1, {'c 1/2': '23.4'})
    [('p0_h.alerts.1.e__1_2.captures.c_1_2', 23.4, 1)]
    >>> _convert_captures('p0.h', 1, 'e1', 1, {'c3': {'c 3.1/': '42'}})
    [('p0_h.alerts.1.e1.captures.c3.c_3_1_', 42.0, 1)]
    '''

    result = []
    key = '{worker_name}.alerts.{alert_id}.{entity_id}.captures.{capture}'

    safe_worker_name = GRAPHITE_REPLACE_KEYCHARS.sub('_', worker_name)
    safe_entity_id = GRAPHITE_REPLACE_KEYCHARS.sub('_', entity_id)

    for capture, value in captures.iteritems():
        safe_capture = GRAPHITE_REPLACE_KEYCHARS.sub('_', capture)
        if isinstance(value, dict):
            for inner_capture, inner_value in value.iteritems():
                try:
                    v = float(inner_value)
                except (ValueError, TypeError):
                    continue
                safe_inner_capture = GRAPHITE_REPLACE_KEYCHARS.sub('_', inner_capture)
                result.append(('{}.{}'.format(key.format(worker_name=safe_worker_name, alert_id=alert_id,
                              entity_id=safe_entity_id, capture=safe_capture), safe_inner_capture), v, timestamp))
        else:
            try:
                v = float(value)
            except (ValueError, TypeError):
                continue
            result.append((key.format(worker_name=safe_worker_name, alert_id=alert_id, entity_id=safe_entity_id,
                          capture=safe_capture), v, timestamp))

    return result



def evaluate_condition(val, condition, **ctx):
    '''

    >>> evaluate_condition(0, '>0')
    False

    >>> evaluate_condition(1, '>0')
    True

    >>> evaluate_condition(1, 'delta("5m")<-10', delta=lambda x:1)
    False

    >>> evaluate_condition({'a': 1}, '["a"]>10')
    False
    '''

    return safe_eval(_prepare_condition(condition), eval_source='<alert-condition>', value=val, **ctx)


class InvalidEvalExpression(Exception):
    pass


class MalformedCheckResult(Exception):

    def __init__(self, msg):
        Exception.__init__(self, msg)


class Try(Callable):

    def __init__(self, try_call, except_call, exc_cls=Exception):
        self.try_call = try_call
        self.except_call = except_call
        self.exc_cls = exc_cls

    def __call__(self, *args):
        try:
            return self.try_call()
        except self.exc_cls, e:
            return self.except_call(e)

def get_results_user(count=1, con=None, check_id=None, entity_id=None):
    return map(lambda x: x["value"], get_results(con, check_id, entity_id, count))

def get_results(con, check_id, entity_id, count=1):
    r = map(json.loads, con.lrange('zmon:checks:{}:{}'.format(check_id, entity_id), 0, count - 1))

    for x in r:
        x.update({"entity_id": entity_id})

    return r


def avg(sequence):
    '''
    >>> avg([])
    0
    >>> avg([1, 2, 3])
    2.0
    >>> avg([2, 3])
    2.5
    '''

    l = len(sequence) * 1.0
    return (sum(sequence) / l if l else 0)


def empty(v):
    '''
    >>> empty([])
    True
    >>> empty([1])
    False
    '''

    return not bool(v)


def build_default_context():
    return {
        'abs': abs,
        'all': all,
        'any': any,
        'avg': avg,
        'basestring': basestring,
        'bin': bin,
        'bool': bool,
        'chain': itertools.chain,
        'chr': chr,
        'Counter': Counter,
        'dict': dict,
        'divmod': divmod,
        'Exception': Exception,
        'empty': empty,
        'enumerate': enumerate,
        'False': False,
        'filter': filter,
        'float': float,
        'groupby': itertools.groupby,
        'hex': hex,
        'int': int,
        'isinstance': isinstance,
        'json': json.loads,
        'len': len,
        'list': list,
        'long': long,
        'map': map,
        'max': max,
        'min': min,
        'normalvariate': random.normalvariate,
        'oct': oct,
        'ord': ord,
        'pow': pow,
        'range': range,
        'reduce': functools.reduce,
        'reversed': reversed,
        'round': round,
        'set': set,
        'sorted': sorted,
        'str': str,
        'sum': sum,
        'timestamp': time.time,
        'True': True,
        'Try': Try,
        'tuple': tuple,
        'unichr': unichr,
        'unicode': unicode,
        'xrange': xrange,
        'zip': zip,
    }


def check_ast_node_is_safe(node):
    '''
    Check that the ast node does not contain any system attribute calls
    as well as exec call (not to construct the system attribute names with strings).

    eval() function calls should not be a problem, as it is hopefuly not exposed
    in the globals and __builtins__

    >>> node = ast.parse('def __call__(): return 1')
    >>> node == check_ast_node_is_safe(node)
    True

    >>> check_ast_node_is_safe(ast.parse('def m(): return ().__class__'))
    Traceback (most recent call last):
        ...
    InvalidEvalExpression: alert definition should not try to access hidden attributes (for example '__class__')


    >>> check_ast_node_is_safe(ast.parse('def horror(g): exec "exploit = ().__" + "class" + "__" in g'))
    Traceback (most recent call last):
        ...
    InvalidEvalExpression: alert definition should not try to execute arbitrary code

    '''

    for n in ast.walk(node):
        if isinstance(n, ast.Attribute):
            if n.attr.startswith('__'):
                raise InvalidEvalExpression("alert definition should not try to access hidden attributes (for example '__class__')"
                                            )
        elif isinstance(n, ast.Exec):
            raise InvalidEvalExpression('alert definition should not try to execute arbitrary code')
    return node


def safe_eval(expr, eval_source='<string>', **kwargs):
    '''
    Safely execute expr.

    For now expr can be only one python expression, a function definition
    or a callable class definition.

    If the expression is returning a callable object (like lambda function
    or Try() object) it will be called and a result of the call will be returned.

    If a result of calling of the defined function or class are returning a callable object
    it will not be called.

    As access to the hidden attributes is protected by check_ast_node_is_safe() method
    we should not have any problem with valnarabilites defined here:
    Link: http://nedbatchelder.com/blog/201206/eval_really_is_dangerous.html

    TODO: implement compile object cache

    >>> safe_eval('value > 0', value=1)
    True

    >>> safe_eval('def m(): return value', value=10)
    10

    >>> safe_eval('def m(param): return value', value=10)
    Traceback (most recent call last):
        ...
    TypeError: m() takes exactly 1 argument (0 given)

    >>> safe_eval('lambda: value', value=10)
    10

    >>> result = safe_eval('def m(): print value', value=10)
    Traceback (most recent call last):
        ...
    SyntaxError: invalid syntax

    >>> result = safe_eval('print value', value=10)
    Traceback (most recent call last):
        ...
    SyntaxError: invalid syntax

    >>> safe_eval('def m(): return lambda: value', value=10) #doctest: +ELLIPSIS
    <function <lambda> at ...>

    >>> safe_eval('error = value', value=10)
    Traceback (most recent call last):
        ...
    InvalidEvalExpression: alert definition can contain a python expression, a function call or a callable class definition

    >>> safe_eval('def m(): return value.__class__', value=10)
    Traceback (most recent call last):
        ...
    InvalidEvalExpression: alert definition should not try to access hidden attributes (for example '__class__')

    >>> safe_eval("""
    ... class CallableClass(object):
    ...
    ...     def get_value(self):
    ...         return value
    ...
    ...     def __call__(self):
    ...         return self.get_value()
    ... """, value=10)
    10

    >>> safe_eval("""
    ... class NotCallableClass(object):
    ...
    ...     def get_value(self):
    ...         return value
    ...
    ...     def call(self): # this is not a callable class
    ...         return self.get_value()
    ... """, value=10)
    Traceback (most recent call last):
        ...
    InvalidEvalExpression: alert definition should contain a callable class definition (missing __call__ method?)


    >>> safe_eval("""
    ... def firstfunc():
    ...     return value
    ...
    ... value > 0
    ...
    ... """, value=10)
    Traceback (most recent call last):
        ...
    InvalidEvalExpression: alert definition should contain only one python expression, a function call or a callable class definition

    '''

    g = {'__builtins__': {}, 'object': object, '__name__': __name__}
    # __builtins__ should be masked away to disable builtin functions
    # object is needed if the NewStyle class is being created
    # __name__ is needed to be able to complie a class
    g.update(kwargs)

    node = compile(expr, eval_source, 'exec', ast.PyCF_ONLY_AST | __future__.CO_FUTURE_PRINT_FUNCTION)
    node = check_ast_node_is_safe(node)
    body = node.body
    if body and len(body) == 1:
        x = body[0]
        if isinstance(x, ast.FunctionDef) or isinstance(x, ast.ClassDef):
            cc = compile(node, eval_source, 'exec')  # can be nicely cached
            v = {}
            exec (cc, g, v)
            if len(v) == 1:
                c = v.itervalues().next()
                if isclass(c):
                    # we need a class instance and not the class itself
                    c = c()

                if callable(c):
                    return c()  # if a function will return another callable, we will not call it
                else:
                    raise InvalidEvalExpression('alert definition should contain a callable class definition (missing __call__ method?)'
                                                )
            else:
                raise InvalidEvalExpression('alert definition should contain only one function or one callable class definition'
                                            )
        elif isinstance(x, ast.Expr):
            cc = compile(expr, eval_source, 'eval', __future__.CO_FUTURE_PRINT_FUNCTION)  # can be nicely cached
            r = eval(cc, g)
            if callable(r):
                # Try() returns callable that should be executed
                return r()
            else:
                return r
        else:
            raise InvalidEvalExpression('alert definition can contain a python expression, a function call or a callable class definition'
                                        )
    else:
        raise InvalidEvalExpression('alert definition should contain only one python expression, a function call or a callable class definition'
                                    )

class NotaZmonTask(object):

    abstract = True
    _host = 'localhost'
    _port = 6379
    _secure_queue = 'zmon:queue:secure'
    _db = 0
    _con = None
    _graphite = None
    _counter = Counter()
    _captures_local = []
    _last_metrics_sent = 0
    _last_captures_sent = 0
    _logger = None
    _logfile = None
    _loglevel = logging.DEBUG
    _zmon_url = None
    _worker_name = None
    _queues = None
    _stash = None
    _stash_cmds = None
    _safe_repositories = []

    _is_secure_worker = True

    _timezone = None
    _account = None
    _team = None
    _dataservice_url = None

    _dataservice_poster = None


    _plugin_category = 'Function'
    _plugins = []
    _function_factories = {}
    _zmon_actuator_checkid = None

    @classmethod
    def configure(cls, config):
        try:
            #configure RedisConnHandler
            RedisConnHandler.configure(**config)
        except KeyError:
            logger.exception('Error creating connection: ')
            raise
        #cls._loglevel = (logging.getLevelName(config['loglevel']) if 'loglevel' in config else logging.INFO)
        cls._logfile = config.get('logfile')
        cls._soap_config = {k: v for k, v in config.items() if k.startswith('soap.service')}
        cls._zmon_url = config.get('zmon.url')
        cls._queues = config.get('zmon.queues', {}).get('local')
        cls._safe_repositories = sorted(config.get('safe_repositories', []))
        cls._zmon_actuator_checkid = config.get('zmon.actuator.checkid', None)

        cls._logger = cls.get_configured_logger()
        cls.perload_stash_commands()

        cls._is_secure_worker = config.get('worker.is_secure')

        cls._timezone = pytz.timezone('Europe/Berlin')

        cls._account = config.get('account')
        cls._team = config.get('team')

        cls._dataservice_url = config.get('dataservice.url')

        if cls._dataservice_url:
            # start action loop for sending reports to dataservice
            cls._logger.info("Enabling data service: {}".format(cls._dataservice_url))
            cls._dataservice_poster = PeriodicBufferedAction(cls.send_to_dataservice, retries=10, t_wait=5)
            cls._dataservice_poster.start()

        cls._plugins = plugin_manager.get_plugins_of_category(cls._plugin_category)
        # store function factories from plugins in a dict by name
        cls._function_factories = {p.name: p.plugin_object for p in cls._plugins}

    def __init__(self):
        self.task_context = None
        self._cmds_first_accessed = False

    @classmethod
    def is_secure_worker(cls):
        return cls._is_secure_worker


    @classmethod
    def perload_stash_commands(cls):
        cls._stash = StashAccessor(cls.get_configured_logger())
        if cls.is_secure_worker():
            try:
                cls._stash_cmds = cls._stash.get_stash_commands(*cls._safe_repositories)
                cls._logger.info('Loaded %d commands from stash secure repos', len(cls._stash_cmds))
            except Exception:
                cls._logger.exception('Error loading stash commands: ')

    @async_memory_cache.cache_on_arguments(namespace='zmon-worker', expiration_time=STASH_CACHE_EXPIRATION_TIME)
    @with_retries(max_retries=3, delay=10)
    def load_stash_commands(self, repositories):
        if not self._cmds_first_accessed:
            # ugly but needed to stop celery from refreshing the cache when task process is forked
            self._cmds_first_accessed = True
            return self._stash_cmds
        else:
            return self._stash.get_stash_commands(*repositories)

    @classmethod
    def get_configured_logger(cls):
        if not cls._logger:
            cls._logger = logger
        return cls._logger

    @property
    def con(self):
        self._con = RedisConnHandler.get_instance().get_conn()
        BaseNotification.set_redis_con(self._con)
        return self._con

    @property
    def logger(self):
        return self.get_configured_logger()

    @property
    def worker_name(self):
        if not self._worker_name:
            self._worker_name = 'p{}.{}'.format('local', socket.gethostname())
        return self._worker_name

    def get_redis_host(self):
        return RedisConnHandler.get_instance().get_parsed_redis().hostname

    def get_redis_port(self):
        return RedisConnHandler.get_instance().get_parsed_redis().port

    def send_metrics(self):
        now = time.time()
        if now > self._last_metrics_sent + METRICS_INTERVAL:
            p = self.con.pipeline()
            p.sadd('zmon:metrics', self.worker_name)
            for key, val in self._counter.items():
                p.incrby('zmon:metrics:{}:{}'.format(self.worker_name, key), val)
            p.set('zmon:metrics:{}:ts'.format(self.worker_name), now)
            p.execute()
            self._counter.clear()
            self._last_metrics_sent = now
            self.logger.info('Send metrics, end storing metrics in redis count: %s, duration: %.3fs', len(self._counter), time.time() - now)

    @classmethod
    def send_to_dataservice(cls, check_results, timeout=10, method='PUT'):

        http_req = {'PUT': requests.put, 'POST': requests.post, 'GET': requests.get}
        headers = {'content-type': 'application/json'}

        team = cls._team if cls._team is not None else ''
        account = cls._account if cls._account is not None else ''

        try:
            # group check_results by check_id
            results_by_id = defaultdict(list)
            for cr in check_results:
                results_by_id[cr['check_id']].append(cr)

            # make separate posts per check_id
            for check_id, results in results_by_id.items():

                url = '{url}/{account}/{check_id}/'.format(url=cls._dataservice_url.rstrip('/'),
                                                           account=urllib.quote(account), check_id=check_id)
                worker_result = {
                    'team': team,
                    'account': account,
                    'results': results,
                }

                r = http_req[method](url, data=json.dumps(worker_result), timeout=timeout, headers=headers)
                if r.status_code != requests.codes.ok:
                    raise Exception('http request to {} got status_code={}'.format(url, r.status_code))

        except Exception:
            logger.exception('Error in dataservice post: ')
            raise


    def check_and_notify(self, req, alerts, task_context=None):
        self.task_context = task_context
        start_time = time.time()
        # soft_time_limit = req['interval']
        check_id = req['check_id']
        entity_id = req['entity']['id']

        try:
            val = self.check(req)
        #TODO: need to support soft and hard time limits soon
        # except SoftTimeLimitExceeded, e:
        #     self.logger.info('Check request with id %s on entity %s exceeded soft time limit', check_id,
        #                                  entity_id)
        #     # PF-3685 It might happen that this exception was raised after sending a command to redis, but before receiving
        #     # a response. In this case, the connection object is "dirty" and when the same connection gets taken out of the
        #     # pool and reused, it'll throw an exception in redis client.
        #     self.con.connection_pool.disconnect()
        #     notify(check_and_notify, {'ts': start_time, 'td': soft_time_limit, 'value': str(e)}, req, alerts,
        #            force_alert=True)
        except CheckError, e:
            # self.logger.warn('Check failed for request with id %s on entity %s. Output: %s', check_id, entity_id, str(e))
            self.notify({'ts': start_time, 'td': time.time() - start_time, 'value': str(e), 'worker': self.worker_name, 'exc': 1}, req, alerts,
                   force_alert=True)
        except SecurityError, e:
            self.logger.exception('Security exception in request with id %s on entity %s', check_id, entity_id)
            self.notify({'ts': start_time, 'td': time.time() - start_time, 'value': str(e), 'worker': self.worker_name, 'exc': 1}, req, alerts,
                   force_alert=True)
        except Exception, e:
            # self.logger.exception('Check request with id %s on entity %s threw an exception', check_id, entity_id)
            # PF-3685 Disconnect on unknown exceptions: we don't know what actually happened, it might be that redis
            # connection is dirty. CheckError exception is "safe", it's thrown by the worker whenever the check returns a
            # different response than expected, the user doesn't have access to the checked entity or there's an error in
            # check's parameters.
            self.con.connection_pool.disconnect()
            self.notify({'ts': start_time, 'td': time.time() - start_time, 'value': str(e), 'worker': self.worker_name, 'exc': 1}, req, alerts,
                   force_alert=True)
        else:
            self.notify(val, req, alerts)

    def trial_run(self, req, alerts, task_context=None):
        self.task_context = task_context
        start_time = time.time()
        # soft_time_limit = req['interval']
        entity_id = req['entity']['id']

        try:
            val = self.check_for_trial_run(req)
        #TODO: need to support soft and hard time limits soon
        # except SoftTimeLimitExceeded, e:
        #     trial_run.logger.info('Trial run on entity %s exceeded soft time limit', entity_id)
        #     trial_run.con.connection_pool.disconnect()
        #     notify_for_trial_run(trial_run, {'ts': start_time, 'td': soft_time_limit, 'value': str(e)}, req, alerts,
        #                          force_alert=True)
        except InsufficientPermissionsError, e:
            self.logger.info('Access denied for user %s to run check on %s', req['created_by'], entity_id)
            eventlog.log(EVENTS['ACCESS_DENIED'].id, userName=req['created_by'], entity=entity_id)
            self.notify_for_trial_run({'ts': start_time, 'td': time.time() - start_time, 'value': str(e)}, req,
                                 alerts, force_alert=True)
        except CheckError, e:
            self.logger.warn('Trial run on entity %s failed. Output: %s', entity_id, str(e))
            self.notify_for_trial_run({'ts': start_time, 'td': time.time() - start_time, 'value': str(e)}, req,
                                 alerts, force_alert=True)
        except Exception, e:
            self.logger.exception('Trial run on entity %s threw an exception', entity_id)
            self.con.connection_pool.disconnect()
            self.notify_for_trial_run({'ts': start_time, 'td': time.time() - start_time, 'value': str(e)}, req,
                                 alerts, force_alert=True)
        else:
            self.notify_for_trial_run(val, req, alerts)


    def cleanup(self, *args, **kwargs):
        self.task_context = kwargs.get('task_context')
        p = self.con.pipeline()
        p.smembers('zmon:checks')
        p.smembers('zmon:alerts')
        check_ids, alert_ids = p.execute()

        for check_id in kwargs.get('disabled_checks', {}):
            self._cleanup_check(p, check_id)

        for alert_id in kwargs.get('disabled_alerts', {}):
            self._cleanup_alert(p, alert_id)

        for check_id in check_ids:
            if check_id in kwargs.get('check_entities', {}):
                redis_entities = self.con.smembers('zmon:checks:{}'.format(check_id))
                check_entities = set(kwargs['check_entities'][check_id])

                # If it happens that we remove all entities for given check, we should remove all the things.
                if not check_entities:
                    p.srem('zmon:checks', check_id)
                    p.delete('zmon:checks:{}'.format(check_id))
                    for entity in redis_entities:
                        p.delete('zmon:checks:{}:{}'.format(check_id, entity))
                else:
                    self._cleanup_common(p, 'checks', check_id, redis_entities - check_entities)
            else:

                self._cleanup_check(p, check_id)

        for alert_id in alert_ids:
            if alert_id in kwargs.get('alert_entities', {}):
                # Entities that are in the alert state.
                redis_entities = self.con.smembers('zmon:alerts:{}'.format(alert_id))
                alert_entities = set(kwargs['alert_entities'][alert_id])

                # If it happens that we remove all entities for given alert, we should remove all the things.
                if not alert_entities:
                    p.srem('zmon:alerts', alert_id)
                    p.delete('zmon:alerts:{}'.format(alert_id))
                    p.delete('zmon:alerts:{}:entities'.format(alert_id))
                    for entity in redis_entities:
                        p.delete('zmon:alerts:{}:{}'.format(alert_id, entity))
                        p.delete('zmon:notifications:{}:{}'.format(alert_id, entity))
                else:
                    self._cleanup_common(p, 'alerts', alert_id, redis_entities - alert_entities)
                    # All entities matching given alert definition.
                    all_entities = set(self.con.hkeys('zmon:alerts:{}:entities'.format(alert_id)))
                    for entity in all_entities - alert_entities:
                        self.logger.info('Removing entity %s from hash %s', entity,
                                            'zmon:alerts:{}:entities'.format(alert_id))
                        p.hdel('zmon:alerts:{}:entities'.format(alert_id), entity)
                        p.delete('zmon:notifications:{}:{}'.format(alert_id, entity))
            else:
                self._cleanup_alert(p, alert_id)

        p.execute()


    def _cleanup_check(self, pipeline, check_id):
        self.logger.info('Removing check with id %s from zmon:checks set', check_id)
        pipeline.srem('zmon:checks', check_id)
        for entity_id in self.con.smembers('zmon:checks:{}'.format(check_id)):
            self.logger.info('Removing key %s', 'zmon:checks:{}:{}'.format(check_id, entity_id))
            pipeline.delete('zmon:checks:{}:{}'.format(check_id, entity_id))
        self.logger.info('Removing key %s', 'zmon:checks:{}'.format(check_id))
        pipeline.delete('zmon:checks:{}'.format(check_id))


    def _cleanup_alert(self, pipeline, alert_id):
        self.logger.info('Removing alert with id %s from zmon:alerts set', alert_id)
        pipeline.srem('zmon:alerts', alert_id)
        for entity_id in self.con.smembers('zmon:alerts:{}'.format(alert_id)):
            self.logger.info('Removing key %s', 'zmon:alerts:{}:{}'.format(alert_id, entity_id))
            pipeline.delete('zmon:alerts:{}:{}'.format(alert_id, entity_id))
            pipeline.delete('zmon:notifications:{}:{}'.format(alert_id, entity_id))
        self.logger.info('Removing key %s', 'zmon:alerts:{}'.format(alert_id))
        pipeline.delete('zmon:alerts:{}'.format(alert_id))
        self.logger.info('Removing key %s', 'zmon:alert:{}:entities'.format(alert_id))
        pipeline.delete('zmon:alerts:{}:entities'.format(alert_id))


    def _cleanup_common(self, pipeline, entry_type, entry_id, entities):
        '''
        Removes entities from redis matching given type and id.
        Parameters
        ----------
        entry_type: str
            Type of entry to remove: 'checks' or 'alerts'.
        entry_id: int
            Id of entry to remove.
        entities: set
            A set of entities to remove (difference between entities from scheduler and ones present in redis).
        '''

        for entity in entities:
            self.logger.info('Removing entity %s from set %s', entity, 'zmon:{}:{}'.format(entry_type, entry_id))
            pipeline.srem('zmon:{}:{}'.format(entry_type, entry_id), entity)
            self.logger.info('Removing key %s', 'zmon:{}:{}:{}'.format(entry_type, entry_id, entity))
            pipeline.delete('zmon:{}:{}:{}'.format(entry_type, entry_id, entity))

    def _store_check_result(self, req, result):
        self.con.sadd('zmon:checks', req['check_id'])
        self.con.sadd('zmon:checks:{}'.format(req['check_id']), req['entity']['id'])
        key = 'zmon:checks:{}:{}'.format(req['check_id'], req['entity']['id'])
        value = json.dumps(result, cls=JsonDataEncoder)
        self.con.lpush(key, value)
        self.con.ltrim(key, 0, DEFAULT_CHECK_RESULTS_HISTORY_LENGTH - 1)


    def check(self, req):

        self.logger.debug(req)
        # schedule_time = req['schedule_time']
        start = time.time()

        try:
            setp(req['check_id'], req['entity']['id'], 'start')
            res = self._get_check_result(req)
            setp(req['check_id'], req['entity']['id'], 'done')
        except Exception, e:
            # PF-3778 Always store check results and re-raise exception which will be handled in 'check_and_notify'.
            self._store_check_result(req, {'td': round(time.time() - start, ROUND_SECONDS_DIGITS), 'ts': round(start,
                                ROUND_SECONDS_DIGITS), 'value': str(e), 'worker': self.worker_name, 'exc': 1})
            raise
        finally:
            # Store duration in milliseconds as redis only supports integers for counters.

            # 'check.{}.count'.format(req['check_id']): 1,
            # 'check.{}.duration'.format(req['check_id']): int(round(1000.0 * (time.time() - start))),
            # 'check.{}.latency'.format(req['check_id']): int(round(1000.0 * (start - schedule_time))),

            self._counter.update({
                'check.count': 1
            })

            self.send_metrics()

        setp(req['check_id'], req['entity']['id'], 'store')
        self._store_check_result(req, res)
        setp(req['check_id'], req['entity']['id'], 'stored')

        return res




    def check_for_trial_run(self, req):
        # fake check ID as it is used by check context
        req['check_id'] = 'trial_run'
        return self._get_check_result(req)


    @timed
    def _get_check_result_internal(self, req):

        self._enforce_security(req)
        cmd = req['command']

        ctx = self._build_check_context(req)
        try:
            result = safe_eval(cmd, eval_source='<check-command>', **ctx)
            return result() if isinstance(result, Callable) else result

        except (SyntaxError, InvalidEvalExpression), e:
            raise CheckError(str(e))

    def _get_check_result(self, req):
        r = self._get_check_result_internal(req)
        r['worker'] = self.worker_name
        return r

    def _enforce_security(self, req):
        '''
        Check tasks from the secure queue to asert the command to run is specified in stash check definition
        Side effect: modifies req to address unique security concerns
        Raises SecurityError on check failure
        '''

        if self.is_secure_worker() or self.task_context['delivery_info'].get('routing_key') == 'secure':
            try:
                stash_commands = self.load_stash_commands(self._safe_repositories)
            except Exception, e:
                traceback = sys.exc_info()[2]
                raise SecurityError('Unexpected Internal error: {}'.format(e)), None, traceback

            if req['command'] not in stash_commands:
                raise SecurityError('Security violation: Non-authorized command received in secure environment')

            # transformations of entities: hostname "pp-whatever" needs to become "whatever.pp"
            prefix = 'pp-'
            if 'host' in req['entity'] and str(req['entity']['host']).startswith(prefix):

                self.logger.warn('secure req[entity] before pp- transformations: %s', req['entity'])

                real_host = req['entity']['host']
                #secure_host = '{}.pp'.format(req['entity']['host'][3:])
                secure_host = '{}.{}'.format(req['entity']['host'][len(prefix):], prefix[:-1])
                # relplace all real host values occurrences with secure_host
                req['entity'].update({k: v.replace(real_host, secure_host) for k, v in req['entity'].items() if
                                      isinstance(v, basestring) and real_host in v and k != 'id'})

                self.logger.warn('secure req[entity] after pp- transformations: %s', req['entity'])



    def _build_check_context(self, req):
        '''Build context for check command with all necessary functions'''

        entity = req['entity']

        # function creation context: passed to function factories create() method
        factory_ctx = {
            'entity': entity,
            'entity_url': _get_entity_url(entity),
            'check_id': req['check_id'],
            'entity_id': entity['id'],
            'host': entity.get('host'),
            'port': entity.get('port'),
            'instance': entity.get('instance'),
            'external_ip': entity.get('external_ip'),
            'load_balancer_status': entity.get('load_balancer_status'),
            'data_center_code': entity.get('data_center_code'),
            'database': entity.get('database'),
            'jmx_port': _get_jmx_port(entity),
            'shards': _get_shards(entity),
            'soft_time_limit': req['interval'],
            'redis_host': self.get_redis_host(),
            'redis_port': self.get_redis_port(),
            'zmon_url': NotaZmonTask._zmon_url,
            'entity_id_for_kairos': normalize_kairos_id(entity['id']),
            'req_created_by': req.get('created_by'),
        }

        # check execution context
        ctx = build_default_context()
        ctx['entity'] = entity

        # populate check context with functions from plugins' function factories
        for func_name, func_factory in self._function_factories.items():
            if func_name not in ctx:
                ctx[func_name] = func_factory.create(factory_ctx)

        return ctx


    def evaluate_alert(self, alert_def, req, result):
        '''Check if the result triggers an alert

        The function will save the global alert state to the following redis keys:

        * zmon:alerts:<ALERT-DEF-ID>:entities    hash of entity IDs -> captures
        * zmon:alerts                            set of active alert definition IDs
        * zmon:alerts:<ALERT-DEF-ID>             set of entity IDs in alert
        * zmon:alerts:<ALERT-DEF-ID>:<ENTITY-ID> JSON with alert evaluation result for given alert definition and entity
        '''

        # captures is our map of "debug" information, e.g. to see values calculated in our condition
        captures = {}

        alert_id = alert_def['id']
        check_id = alert_def['check_id']
        alert_parameters = alert_def.get('parameters')

        try:
            result = evaluate_condition(result['value'], alert_def['condition'], **build_condition_context(self.con,
                                        check_id, alert_id, req['entity'], captures, alert_parameters))
        except Exception, e:
            captures['exception'] = str(e)
            result = True

        try:
            is_alert = bool((result() if isinstance(result, Callable) else result))
        except Exception, e:
            captures['exception'] = str(e)
            is_alert = True

        # add parameters to captures so they can be substituted in alert title
        if alert_parameters:
            pure_captures = captures.copy()
            try:
                captures = {k: p['value'] for k, p in alert_parameters.items()}
            except Exception, e:
                self.logger.exception('Error when capturing parameters: ')
            captures.update(pure_captures)

        return is_alert, captures


    def send_notification(self, notification, context):
        ctx = _build_notify_context(context)
        try:
            repeat = safe_eval(notification, eval_source='<check-command>' , **ctx)
        except Exception, e:
            # TODO Define what should happen if sending emails or sms fails.
            self.logger.exception(e)
        else:
            if repeat:
                self.con.hset('zmon:notifications:{}:{}'.format(context['alert_def']['id'], context['entity']['id']),
                              notification, time.time() + repeat)

    def notify(self, val, req, alerts, force_alert=False):
        '''
        Process check result and evaluate all alerts. Returns list of active alert IDs
        Parameters
        ----------
        val: dict
            Check result, see check function
        req: dict
            Check request dict
        alerts: list
            A list of alert definitions matching the checked entity
        force_alert: bool
            An optional flag whether to skip alert evalution and force "in alert" state. Used when check request exceeds
            time limit or throws other exception, this way unexpected conditions are always treated as alerts.
        Returns
        -------
        list
            A list of alert definitions matching given entity.
        '''

        ts_serialize = lambda ts: datetime.fromtimestamp(ts, tz=self._timezone).isoformat(' ') if ts else None

        result = []
        entity_id = req['entity']['id']
        start = time.time()

        check_result = {
            'time': ts_serialize(val.get('ts')) if isinstance(val, dict) else None,
            'run_time': val.get('td') if isinstance(val, dict) else None,  # TODO: should be float or is it milliseconds?
            'check_id': req['check_id'],
            'entity_id': req['entity']['id'],
            'check_result': val,
            'exception': True if isinstance(val, dict) and val.get('exc') else False,
            'alerts': {},
        }

        try:
            setp(req['check_id'], entity_id, 'notify loop')
            for alert in alerts:
                alert_id = alert['id']
                alert_entities_key = 'zmon:alerts:{}'.format(alert_id)
                alerts_key = 'zmon:alerts:{}:{}'.format(alert_id, entity_id)
                notifications_key = 'zmon:notifications:{}:{}'.format(alert_id, entity_id)
                is_alert, captures = ((True, {}) if force_alert else self.evaluate_alert(alert, req, val))

                func = getattr(self.con, ('sadd' if is_alert else 'srem'))
                changed = bool(func(alert_entities_key, entity_id))

                if is_alert:
                    # bubble up: also update global set of alerts
                    alert_changed = func('zmon:alerts', alert_id)

                    if alert_changed:
                        _log_event('ALERT_STARTED', alert, val)
                else:
                    entities_in_alert = self.con.smembers(alert_entities_key)
                    if not entities_in_alert:
                        # no entity has alert => remove from global set
                        alert_changed = func('zmon:alerts', alert_id)
                        if alert_changed:
                            _log_event('ALERT_ENDED', alert, val)

                # PF-3318 If an alert has malformed time period, we should evaluate it anyway and continue with
                # the remaining alert definitions.
                try:
                    is_in_period = in_period(alert.get('period', ''))
                except InvalidFormat, e:
                    self.logger.warn('Alert with id %s has malformed time period.', alert_id)
                    captures['exception'] = '; \n'.join(filter(None, [captures.get('exception'), str(e)]))
                    is_in_period = True

                if changed and is_in_period and is_alert:
                    # notify on entity-level
                    _log_event(('ALERT_ENTITY_STARTED'), alert, val, entity_id)
                elif changed and not is_alert:
                    _log_event(('ALERT_ENTITY_ENDED'), alert, val, entity_id)


                # Always store captures for given alert-entity pair, this is also used a list of all entities matching
                # given alert id. Captures are stored here because this way we can easily link them with check results
                # (see PF-3146).
                self.con.hset('zmon:alerts:{}:entities'.format(alert_id), entity_id, json.dumps(captures,
                              cls=JsonDataEncoder))

                # prepare report - alert part
                check_result['alerts'][alert_id] = {
                    'alert_id': alert_id,
                    'captures': captures,
                    'downtimes': [],
                    'exception': True if isinstance(captures, dict) and 'exception' in captures else False,
                    'active':  is_alert,
                    'changed': changed,
                    'in_period': is_in_period,
                    'start_time': None,
                    # '_alert_stored': None,
                    # '_notifications_stored': None,
                }

                # get last alert data stored in redis if any
                alert_stored = None
                try:
                    stored_raw = self.con.get(alerts_key)
                    alert_stored = json.loads(stored_raw) if stored_raw else None
                except (ValueError, TypeError):
                    self.logger.warn('My messy Error parsing JSON alert result for key: %s', alerts_key)

                if False:
                    # get notification data stored in redis if any
                    notifications_stored = None
                    try:
                        stored_raw = self.con.get(notifications_key)
                        notifications_stored = json.loads(stored_raw) if stored_raw else None
                    except (ValueError, TypeError):
                        self.logger.warn('My requete-messy Error parsing JSON alert result for key: %s', notifications_key)

                downtimes = None

                if is_in_period:

                    self._counter.update({'alerts.{}.count'.format(alert_id): 1,
                                         'alerts.{}.evaluation_duration'.format(alert_id): int(round(1000.0 * (time.time()
                                         - start)))})

                    # Always evaluate downtimes, so that we don't miss downtime_ended event in case the downtime ends when
                    # the alert is no longer active.
                    downtimes = self._evaluate_downtimes(alert_id, entity_id)

                    start_time = time.time()

                    # Store or remove the check value that triggered the alert
                    if is_alert:
                        result.append(alert_id)
                        start_time = alert_stored['start_time'] if alert_stored and not changed else time.time()

                        # create or refresh stored alert
                        alert_stored = dict(captures=captures, downtimes=downtimes, start_time=start_time, **val)
                        self.con.set(alerts_key, json.dumps(alert_stored, cls=JsonDataEncoder))
                    else:
                        self.con.delete(alerts_key)
                        self.con.delete(notifications_key)

                    start = time.time()
                    notification_context = {
                        'alert_def': alert,
                        'entity': req['entity'],
                        'value': val,
                        'captures': captures,
                        'worker': self.worker_name,
                        'is_alert': is_alert,
                        'changed': changed,
                        'duration': timedelta(seconds=(time.time() - start_time if is_alert and not changed else 0)),
                    }

                    #do not send notifications for downtimed alerts
                    if not downtimes:
                        if changed:
                           for notification in alert['notifications']:
                                self.send_notification(notification, notification_context)
                        else:
                            previous_times = self.con.hgetall(notifications_key)
                            for notification in alert['notifications']:
                                if notification in previous_times and time.time() > float(previous_times[notification]):
                                    self.send_notification(notification, notification_context)

                    self._counter.update({'alerts.{}.notification_duration'.format(alert_id): int(round(1000.0
                                         * (time.time() - start)))})
                    setp(req['check_id'], entity_id, 'notify loop - send metrics')
                    self.send_metrics()
                    setp(req['check_id'], entity_id, 'notify loop end')
                else:
                    self.logger.debug('Alert %s is not in time period: %s', alert_id, alert['period'])
                    if is_alert:
                        entities_in_alert = self.con.smembers('zmon:alerts:{}'.format(alert_id))

                        p = self.con.pipeline()
                        p.srem('zmon:alerts:{}'.format(alert_id), entity_id)
                        p.delete('zmon:alerts:{}:{}'.format(alert_id, entity_id))
                        p.delete(notifications_key)
                        if len(entities_in_alert) == 1:
                            p.srem('zmon:alerts', alert_id)
                        p.execute()

                        self.logger.info('Removed alert with id %s on entity %s from active alerts due to time period: %s',
                                         alert_id, entity_id, alert.get('period', ''))

                # add to alert report regardless alert up/down/out of period
                # report['results']['alerts'][alert_id]['_alert_stored'] = alert_stored
                # report['results']['alerts'][alert_id]['_notifications_stored'] = notifications_stored

                check_result['alerts'][alert_id]['start_time'] = ts_serialize(alert_stored['start_time']) if alert_stored else None
                check_result['alerts'][alert_id]['start_time_ts'] = alert_stored['start_time'] if alert_stored else None
                check_result['alerts'][alert_id]['downtimes'] = downtimes

            setp(req['check_id'], entity_id, 'return notified')

            # enqueue report to be sent via http request
            if self._dataservice_poster:
                #'entity_id': req['entity']['id'],
                check_result["entity"] = {"id": req['entity']['id']}

                for k in ["application_id","application_version","stack_name","stack_version","team","account_alias"]:
                    if k in req["entity"]:
                        check_result["entity"][k] = req["entity"][k]

                self._dataservice_poster.enqueue(check_result)

            return result
        #TODO: except SoftTimeLimitExceeded:
        except Exception:
            # Notifications should not exceed the time limit.
            self.logger.exception('Notification for check %s reached soft time limit', req['check_name'])
            self.con.connection_pool.disconnect()
            return None


    def post_trial_run(self, id, entity, result):
        if self._dataservice_url is not None:

            val = {
                'id': id,
                'entity-id': entity,
                'result': result
            }

            try:
                requests.put(self._dataservice_url+"trial-run/", data=json.dumps(val, cls=JsonDataEncoder), headers={"Content-Type":"application/json"})
            except Exception as ex:
                self.logger.exception(ex)


    def notify_for_trial_run(self, val, req, alerts, force_alert=False):
        """Like notify(), but for trial runs!"""

        try:
            # There must be exactly one alert in alerts.
            alert,  = alerts
            redis_key = 'zmon:trial_run:{uuid}:results'.format(uuid=(alert['id'])[3:])

            is_alert, captures = ((True, {}) if force_alert else self.evaluate_alert(alert, req, val))

            try:
                is_in_period = in_period(alert.get('period', ''))
            except InvalidFormat, e:
                self.logger.warn('Alert with id %s has malformed time period.', alert['id'])
                captures['exception'] = '; \n'.join(filter(None, [captures.get('exception'), str(e)]))
                is_in_period = True

            try:
                result = {
                    'entity': req['entity'],
                    'value': val,
                    'captures': captures,
                    'is_alert': is_alert,
                    'in_period': is_in_period,
                    }
                result_json = json.dumps(result, cls=JsonDataEncoder)
            except TypeError, e:
                result = {
                    'entity': req['entity'],
                    'value': str(e),
                    'captures': {},
                    'is_alert': is_alert,
                    'in_period': is_in_period,
                    }
                result_json = json.dumps(result, cls=JsonDataEncoder)

            self.con.hset(redis_key, req['entity']['id'], result_json)
            self.con.expire(redis_key, TRIAL_RUN_RESULT_EXPIRY_TIME)

            self.post_trial_run(alert['id'][3:], req['entity'], result)

            return ([alert['id']] if is_alert and is_in_period else [])

        #TODO: except SoftTimeLimitExceeded:
        except Exception:
            self.con.connection_pool.disconnect()
            return None


    def _store_captures_locally(self, alert_id, entity_id, timestamp, captures):
        metrics = _convert_captures(self.worker_name, alert_id, entity_id, timestamp, captures)
        if metrics:
            self._captures_local.extend(metrics)



    def _evaluate_downtimes(self, alert_id, entity_id):
        result = []

        p = self.con.pipeline()
        p.smembers('zmon:downtimes:{}'.format(alert_id))
        p.hgetall('zmon:downtimes:{}:{}'.format(alert_id, entity_id))
        redis_entities, redis_downtimes = p.execute()

        try:
            downtimes = dict((k, json.loads(v)) for (k, v) in redis_downtimes.iteritems())
        except ValueError, e:
            self.logger.exception(e)
        else:
            now = time.time()
            for uuid, d in downtimes.iteritems():
                # PF-3604 First check if downtime is active, otherwise check if it's expired, else: it's a future downtime.
                if now > d['start_time'] and now < d['end_time']:
                    d['id'] = uuid
                    result.append(d)
                    func = 'sadd'
                elif now >= d['end_time']:
                    func = 'srem'
                else:
                    continue

                # Check whether the downtime changed state: active -> inactive or inactive -> active.
                changed = getattr(self.con, func)('zmon:active_downtimes', '{}:{}:{}'.format(alert_id, entity_id, uuid))
                if changed:
                    eventlog.log(EVENTS[('DOWNTIME_ENDED' if func == 'srem' else 'DOWNTIME_STARTED')].id, **{
                        'alertId': alert_id,
                        'entity': entity_id,
                        'startTime': d['start_time'],
                        'endTime': d['end_time'],
                        'userName': d['created_by'],
                        'comment': d['comment'],
                    })

                # If downtime is over, we can remove its definition from redis.
                if func == 'srem':
                    if len(downtimes) == 1:
                        p.delete('zmon:downtimes:{}:{}'.format(alert_id, entity_id))
                        if len(redis_entities) == 1:
                            p.delete('zmon:downtimes:{}'.format(alert_id))
                            p.srem('zmon:downtimes', alert_id)
                        else:
                            p.srem('zmon:downtimes:{}'.format(alert_id), entity_id)
                    else:
                        p.hdel('zmon:downtimes:{}:{}'.format(alert_id, entity_id), uuid)
                    p.execute()

        return result
